"""Test case to verify the CAF/CAP calibration"""
import unittest
from pathlib import Path

from pyscos2000 import SCOS, PacketInterface, Parameter


class TestablePacket(PacketInterface):
    """A packet type used for unit testing"""
    def __init__(self, scos, spid, parameters=None):
        super().__init__()
        pids = [r for r in scos.pid if r['PID_SPID'].value == spid]
        assert len(pids) == 1
        self.pid = pids[0]
        self.parameters = {}
        if parameters is not None:
            for name, value in parameters:
                pcf = self.pid.scos.pcf.get(name)
                self.parameters[name] = TestableParameter(value, self, pcf)

    def get(self, item):
        return self.parameters[item]

    def get_all(self, item):
        return [self.parameters[item]]

    def has(self, item):
        return item in self.parameters

    def __len__(self):
        return len(self.parameters)

    def iter(self):
        for value in self.parameters.values():
            yield value


class TestableParameter(Parameter):
    """A packet parameter type used for testing"""
    def __init__(self, value, packet, pcf):
        super().__init__(packet, pcf)
        assert isinstance(value, bytes)
        self.value = value

    def get_raw(self):
        return self.paramtype.unpack(self.value)

    def get_hex(self):
        return self.value.hex()


class CAPCalibTest(unittest.TestCase):
    def setUp(self):
        self.scos = SCOS(Path(__file__).parent / 'capcalib')
        self.scos.check()

    def test_simple1(self):
        """Test simple interpolation"""
        packet = TestablePacket(self.scos, 15,
                                [('PCF01', b'\x05')])
        self.assertEqual(packet['PCF01'].raw, 5)
        self.assertEqual(packet['PCF01'].eng, 2.5)

    def test_simple2(self):
        """Test simple interpolation"""
        packet = TestablePacket(self.scos, 15,
                                [('PCF01', b'\x0b')])
        self.assertEqual(packet['PCF01'].raw, 11)
        self.assertEqual(packet['PCF01'].eng, 4.5)

    def test_exact1(self):
        """Test an exactly matching point"""
        packet = TestablePacket(self.scos, 15,
                                [('PCF01', b'\x08')])
        self.assertEqual(packet['PCF01'].raw, 8)
        self.assertEqual(packet['PCF01'].eng, 2)

    def test_exact2(self):
        """Test an exactly matching point"""
        packet = TestablePacket(self.scos, 15,
                                [('PCF01', b'\x01')])
        self.assertEqual(packet['PCF01'].raw, 1)
        self.assertEqual(packet['PCF01'].eng, 0)

    def test_extrapolate_fail1(self):
        """Test a failing (because not defined) extrapolation"""
        packet = TestablePacket(self.scos, 15,
                                [('PCF01', b'\x0f')])
        self.assertEqual(packet['PCF01'].raw, 15)
        self.assertEqual(packet['PCF01'].eng, None)

    def test_extrapolate_fail2(self):
        """Test a failing (because not defined) extrapolation"""
        packet = TestablePacket(self.scos, 15,
                                [('PCF01', b'\x00')])
        self.assertEqual(packet['PCF01'].raw, 0)
        self.assertEqual(packet['PCF01'].eng, None)

    def test_extrapolate1(self):
        """Test a succeeding extrapolation (beyond upper)"""
        packet = TestablePacket(self.scos, 15,
                                [('PCF02', b'\x0f')])
        self.assertEqual(packet['PCF02'].raw, 15)
        self.assertEqual(packet['PCF02'].eng, 2.5)

    def test_extrapolate2(self):
        """Test a succeeding extrapolation (below lower)"""
        packet = TestablePacket(self.scos, 15,
                                [('PCF02', b'\x00')])
        self.assertEqual(packet['PCF02'].raw, 0)
        self.assertTrue(abs(packet['PCF02'].eng + 1.6666) < 0.0001)


"""The definition of an SCOS-2000 table schema

Based on EGOS-MCS-S2K-ICD-0001 issue 7.2-FINAL (2019-12-13)
"""
import io
import string
from pathlib import Path
from zipfile import ZipFile

from .instance import Table, TableCell
from .checker import SCOSTableChecker, ALL_CHECKERS
from .reporter import StdErrReporter
from .schema import (
    TableDefinition, ColumnDefinition,
    Number, Boolean, String, ParameterNature,
    CalibrationCategoryColumn, EventPacketIndicator,
    EndianessColumn, OutputViewColumn, InterpolationColumn,
    ParameterNatureColumn, JustifyColumn, NumericFormatTypeColumn,
    RadixColumn, LimitInterpretationColumn, LimitTypeInterpretationColumn,
    MonitoringCheckFlagColumn, PacketHeaderElementTypeColumn, PlannableColumn,
    CommandVerificationColumn, CommandInputFormatColumn, CommandCategoryColumn,
    CommandInterlockColumn, DisplayFormatColumn, IntegerFormatTypeColumn,
    TimeTagTypeColumn, PlannableBySourceColumn, StandaloneExecutionColumn,
    SequenceElementTypeColumn, ReleaseTimeTypeColumn, RelativePositionColumn,
    CommandSequenceInterlockColumn, ValueEditableColumn, CommandElementTypeColumn,
    ValueSourceColumn, FormatParameterTypeColumn, CommandValueRepresentationColumn,
    CommandElementType, CommandValueRepresentation, Radix,
    )
from .parametertypes import ParameterType, SignedIntegerParameter, UnsignedIntegerParameter
from .synthetic import DynamicSyntheticParameter


class SyntheticImplementationMissingError(RuntimeError):
    """Raised when the implementation of a synthetic parameter is missing"""


class VDFTable(TableDefinition):
    """The vdf.dat table in SCOS

    Database meta and version information"""
    def __init__(self):
        super().__init__('vdf', [
            ColumnDefinition('VDF_NAME', String(8), optional=False),
            ColumnDefinition('VDF_COMMENT', String(32), optional=True),
            ColumnDefinition('VDF_DOMAIN', Number(5), optional=True),
            ColumnDefinition('VDF_RELEASE', Number(5), optional=True, default='0'),
            ColumnDefinition('VDF_ISSUE', Number(5), optional=True, default='0'),
            ])


class PCFTable(TableDefinition):
    """The pcf.dat table in SCOS

    Monitoring parameters characteristics"""
    def __init__(self):
        super().__init__('pcf', [
            ColumnDefinition('PCF_NAME', String(8), optional=False, is_key=True),
            ColumnDefinition('PCF_DESCR', String(24), optional=True),
            ColumnDefinition('PCF_PID', Number(10), optional=True),
            ColumnDefinition('PCF_UNIT', String(4), optional=True),
            ColumnDefinition('PCF_PTC', Number(2), optional=False),
            ColumnDefinition('PCF_PFC', Number(5), optional=False),
            ColumnDefinition('PCF_WIDTH', Number(6), optional=True),
            ColumnDefinition('PCF_VALID', String(8), optional=True),
            ColumnDefinition('PCF_RELATED', String(8), optional=True),
            ColumnDefinition('PCF_CATEG', CalibrationCategoryColumn(), optional=False),
            ColumnDefinition('PCF_NATUR', ParameterNatureColumn(), optional=False),
            ColumnDefinition('PCF_CURTX', String(10), optional=True),
            ColumnDefinition('PCF_INTER', InterpolationColumn(), optional=True, default='F'),
            ColumnDefinition('PCF_USCON', Boolean(), optional=True, default='N'),
            ColumnDefinition('PCF_DECIM', Number(3), optional=True),
            ColumnDefinition('PCF_PARVAL', String(14), optional=True),
            ColumnDefinition('PCF_SUBSYS', String(8), optional=True),
            ColumnDefinition('PCF_VALPAR', Number(5), optional=True, default='1'),
            ColumnDefinition('PCF_SPTYPE', OutputViewColumn(), optional=True),
            ColumnDefinition('PCF_CORR', Boolean(), optional=True, default='Y'),
            ColumnDefinition('PCF_OBTID', Number(5), optional=True),
            ColumnDefinition('PCF_DARC', Boolean('1', '0'), optional=True, default='0'),
            ColumnDefinition('PCF_ENDIAN', EndianessColumn(), optional=True, default='B'),
            ColumnDefinition('PCF_DESCR2', String(256), optional=True, default=''),
            ])

    def is_synthetic(self, value):
        """Whether or not ``value`` (of type ``ParameterNature``) is synthetic"""
        return value in [ParameterNature.SYNTHETIC,
                         ParameterNature.DYNAMIC,
                         ParameterNature.HARDCODED,
                         ParameterNature.SAVEDSYNTHETIC]


class PIDTable(TableDefinition):
    """The pid.dat table in SCOS

    Packet identification"""
    def __init__(self):
        super().__init__('pid', [
            ColumnDefinition('PID_TYPE', Number(3), optional=False, is_key=True),
            ColumnDefinition('PID_STYPE', Number(3), optional=False, is_key=True),
            ColumnDefinition('PID_APID', Number(5), optional=False, is_key=True),
            ColumnDefinition('PID_PI1_VAL', Number(10), optional=True, default='0', is_key=True),
            ColumnDefinition('PID_PI2_VAL', Number(10), optional=True, default='0', is_key=True),
            ColumnDefinition('PID_SPID', Number(10), optional=False, is_key=True),
            ColumnDefinition('PID_DESCR', String(64), optional=True),
            ColumnDefinition('PID_UNIT', String(8), optional=True),
            ColumnDefinition('PID_TPSD', Number(10), optional=True, default='-1'),
            ColumnDefinition('PID_DFHSIZE', Number(2), optional=False),
            ColumnDefinition('PID_TIME', Boolean(), optional=True, default='N'),
            ColumnDefinition('PID_INTER', Number(10), optional=True),
            ColumnDefinition('PID_VALID', Boolean(), optional=True, default='Y'),
            ColumnDefinition('PID_CHECK', Number(1), optional=True, default='0'),
            ColumnDefinition('PID_EVENT', EventPacketIndicator(), optional=True, default='N'),
            ColumnDefinition('PID_EVID', String(17), optional=True),
            ])


class PICTable(TableDefinition):
    """The pic.dat table in SCOS

    Packet identification criteria"""
    NONE_APID = 99999

    def __init__(self):
        super().__init__('pic', [
            ColumnDefinition('PIC_TYPE', Number(3), optional=False, is_key=True),
            ColumnDefinition('PIC_STYPE', Number(3), optional=False, is_key=True),
            ColumnDefinition('PIC_PI1_OFF', Number(5), optional=False),
            ColumnDefinition('PIC_PI1_WID', Number(3), optional=False),
            ColumnDefinition('PIC_PI2_OFF', Number(5), optional=False),
            ColumnDefinition('PIC_PI2_WID', Number(3), optional=False),
            ColumnDefinition('PIC_APID', Number(5), optional=True, is_key=True,
                             default=str(PICTable.NONE_APID)),
            ])


class TPCFTable(TableDefinition):
    """The tpcf.dat table in SCOS

    Telemetry packets characteristics"""
    def __init__(self):
        super().__init__('tpcf', [
            ColumnDefinition('TPCF_SPID', Number(10), optional=False, is_key=True),
            ColumnDefinition('TPCF_NAME', String(12), optional=True),
            ColumnDefinition('TPCF_SIZE', Number(8), optional=True),
            ])


class PLFTable(TableDefinition):
    """The plf.dat table in SCOS

    Parameter location in fixed packets"""
    def __init__(self):
        super().__init__('plf', [
            ColumnDefinition('PLF_NAME', String(8), optional=False, is_key=True),
            ColumnDefinition('PLF_SPID', Number(10), optional=False, is_key=True),
            ColumnDefinition('PLF_OFFBY', Number(5), optional=False),
            ColumnDefinition('PLF_OFFBI', Number(1), optional=False),
            ColumnDefinition('PLF_NBOCC', Number(4), optional=True, default='1'),
            ColumnDefinition('PLF_LGOCC', Number(5), optional=True, default='0'),
            ColumnDefinition('PLF_TIME', Number(9), optional=True, default='0'),
            ColumnDefinition('PLF_TDOCC', Number(9), optional=True, default='1'),
            ])


class CURTable(TableDefinition):
    """The cur.dat table in SCOS

    Calibration conditional selection"""
    def __init__(self):
        super().__init__('cur', [
            ColumnDefinition('CUR_PNAME', String(8), optional=False, is_key=True),
            ColumnDefinition('CUR_POS', Number(2), optional=False, is_key=True),
            ColumnDefinition('CUR_RLCHK', String(8), optional=False),
            ColumnDefinition('CUR_VALPAR', Number(5), optional=False),
            ColumnDefinition('CUR_SELECT', String(10), optional=False),
            ])


class CAFTable(TableDefinition):
    """The caf.dat table in SCOS

    Numerical calibrations"""
    def __init__(self):
        super().__init__('caf', [
            ColumnDefinition('CAF_NUMBR', String(10), optional=False, is_key=True),
            ColumnDefinition('CAF_DESCR', String(32), optional=True),
            ColumnDefinition('CAF_ENGFMT', NumericFormatTypeColumn(), optional=False),
            ColumnDefinition('CAF_RAWFMT', NumericFormatTypeColumn(), optional=False),
            ColumnDefinition('CAF_RADIX', RadixColumn(), optional=True),
            ColumnDefinition('CAF_UNIT', String(4), optional=True),
            ColumnDefinition('CAF_NCURVE', Number(3), optional=True),
            ColumnDefinition('CAF_INTER', InterpolationColumn(), optional=True, default='F'),
            ])


class CAPTable(TableDefinition):
    """The cap.dat table in SCOS

    Numerical calibration definitions"""
    def __init__(self):
        super().__init__('cap', [
            ColumnDefinition('CAP_NUMBR', String(10), optional=False, is_key=True),
            ColumnDefinition('CAP_XVALS', String(14), optional=False, is_key=True),
            ColumnDefinition('CAP_YVALS', String(14), optional=False),
            ])


class TXFTable(TableDefinition):
    """The txf.dat table in SCOS

    Textual calibrations"""
    def __init__(self):
        super().__init__('txf', [
            ColumnDefinition('TXF_NUMBR', String(10), optional=False, is_key=True),
            ColumnDefinition('TXF_DESCR', String(32), optional=True),
            ColumnDefinition('TXF_RAWFMT', NumericFormatTypeColumn(), optional=False),
            ColumnDefinition('TXF_NALIAS', Number(3), optional=True),
            ])


class TXPTable(TableDefinition):
    """The txp.dat table in SCOS

    Textual calibrations definition"""
    def __init__(self):
        super().__init__('txp', [
            ColumnDefinition('TXP_NUMBR', String(10), optional=False, is_key=True),
            ColumnDefinition('TXP_FROM', String(14), optional=False, is_key=True),
            ColumnDefinition('TXP_TO', String(14), optional=False),
            ColumnDefinition('TXP_ALTXT', String(14), optional=False),
            ])


class MCFTable(TableDefinition):
    """The mcf.dat table in SCOS

    Polynomial calibrations definitions"""
    def __init__(self):
        super().__init__('mcf', [
            ColumnDefinition('MCF_IDENT', String(10), optional=False, is_key=True),
            ColumnDefinition('MCF_DESCR', String(32), optional=True),
            ColumnDefinition('MCF_POL1', String(14), optional=False),
            ColumnDefinition('MCF_POL2', String(14), optional=True, default='0'),
            ColumnDefinition('MCF_POL3', String(14), optional=True, default='0'),
            ColumnDefinition('MCF_POL4', String(14), optional=True, default='0'),
            ColumnDefinition('MCF_POL5', String(14), optional=True, default='0'),
            ])


class LGFTable(TableDefinition):
    """The lgf.dat table in SCOS

    Logarithmic calibrations definitions"""
    def __init__(self):
        super().__init__('lgf', [
            ColumnDefinition('LGF_IDENT', String(10), optional=False, is_key=True),
            ColumnDefinition('LGF_DESCR', String(32), optional=True),
            ColumnDefinition('LGF_POL1', String(14), optional=False),
            ColumnDefinition('LGF_POL2', String(14), optional=True, default='0'),
            ColumnDefinition('LGF_POL3', String(14), optional=True, default='0'),
            ColumnDefinition('LGF_POL4', String(14), optional=True, default='0'),
            ColumnDefinition('LGF_POL5', String(14), optional=True, default='0'),
            ])


class OCFTable(TableDefinition):
    """The ocf.dat table in SCOS

    Monitoring checks"""
    def __init__(self):
        super().__init__('ocf', [
            ColumnDefinition('OCF_NAME', String(8), optional=False, is_key=True),
            ColumnDefinition('OCF_NBCHCK', Number(2), optional=False),
            ColumnDefinition('OCF_NBOOL', Number(2), optional=False),
            ColumnDefinition('OCF_INTER', LimitInterpretationColumn(), optional=False),
            ColumnDefinition('OCF_CODIN', LimitTypeInterpretationColumn(), optional=False),
            ])


class OCPTable(TableDefinition):
    """the ocp.dat table in SCOS

    Monitoring checks definition"""
    def __init__(self):
        super().__init__('ocp', [
            ColumnDefinition('OCP_NAME', String(8), optional=False, is_key=True),
            ColumnDefinition('OCP_POS', Number(2), optional=False, is_key=True),
            ColumnDefinition('OCP_TYPE', MonitoringCheckFlagColumn(), optional=False),
            ColumnDefinition('OCP_LVALU', String(14), optional=True),
            ColumnDefinition('OCP_HVALU', String(14), optional=True),
            ColumnDefinition('OCP_RLCHK', String(8), optional=True),
            ColumnDefinition('OCP_VALPAR', Number(5), optional=True, default='1'),
            ])


class VPDTable(TableDefinition):
    """The vpd.dat table in SCOS

    Variable packet definition"""
    def __init__(self):
        super().__init__('vpd', [
            ColumnDefinition('VPD_TPSD', Number(10), optional=False, is_key=True),
            ColumnDefinition('VPD_POS', Number(4), optional=False, is_key=True),
            ColumnDefinition('VPD_NAME', String(8), optional=False),
            ColumnDefinition('VPD_GRPSIZE', Number(3), optional=True, default='0'),
            ColumnDefinition('VPD_FIXREP', Number(3), optional=True, default='0'),
            ColumnDefinition('VPD_CHOICE', Boolean(), optional=True, default='N'),
            ColumnDefinition('VPD_PIDREF', Boolean(), optional=True, default='N'),
            ColumnDefinition('VPD_DISDESC', String(16), optional=True),
            ColumnDefinition('VPD_WIDTH', Number(2), optional=False),
            ColumnDefinition('VPD_JUSTIFY', JustifyColumn(), optional=True, default='L'),
            ColumnDefinition('VPD_NEWLINE', Boolean(), optional=True, default='N'),
            ColumnDefinition('VPD_DCHAR', Number(1), optional=True, default='0'),
            ColumnDefinition('VPD_FORM', DisplayFormatColumn(), optional=True, default='N'),
            ColumnDefinition('VPD_OFFSET', Number(6), optional=True, default='0'),
            ])


class TCPTable(TableDefinition):
    """The tcp.dat table in SCOS

    TC packet header characteristics"""
    def __init__(self):
        super().__init__('tcp', [
            ColumnDefinition('TCP_ID', String(8), optional=False, is_key=True),
            ColumnDefinition('TCP_DESC', String(24), optional=True),
            ])


class PCPCTable(TableDefinition):
    """The pcpc.dat table in SCOS

    TC packet header parameters"""
    def __init__(self):
        super().__init__('pcpc', [
            ColumnDefinition('PCPC_PNAME', String(8), optional=False, is_key=True),
            ColumnDefinition('PCPC_DESC', String(24), optional=False),
            ColumnDefinition('PCPC_CODE', IntegerFormatTypeColumn(), optional=True, default='U'),
            ])


class PCDFTable(TableDefinition):
    """The pcdf.dat table in SCOS

    Packet heaer definitions"""
    def __init__(self):
        super().__init__('pcdf', [
            ColumnDefinition('PCDF_TCNAME', String(8), optional=False, is_key=True),
            ColumnDefinition('PCDF_DESC', String(24), optional=True),
            ColumnDefinition('PCDF_TYPE', PacketHeaderElementTypeColumn(), optional=False),
            ColumnDefinition('PCDF_LEN', Number(4), optional=False),
            ColumnDefinition('PCDF_BIT', Number(4), optional=False, is_key=True),
            ColumnDefinition('PCDF_PNAME', String(8), optional=True),
            ColumnDefinition('PCDF_VALUE', String(10), optional=False),
            ColumnDefinition('PCDF_RADIX', RadixColumn(), optional=True, default='H'),
            ])


class CCFTable(TableDefinition):
    """The ccf.dat table in SCOS

    Command characteristics"""
    def __init__(self):
        super().__init__('ccf', [
            ColumnDefinition('CCF_CNAME', String(8), optional=False, is_key=True),
            ColumnDefinition('CCF_DESCR', String(24), optional=False),
            ColumnDefinition('CCF_DESCR2', String(64), optional=True),
            ColumnDefinition('CCF_CTYPE', String(8), optional=True),
            ColumnDefinition('CCF_CRITICAL', Boolean(), optional=True, default='N'),
            ColumnDefinition('CCF_PKTID', String(8), optional=False),
            ColumnDefinition('CCF_TYPE', Number(3), optional=True),
            ColumnDefinition('CCF_STYPE', Number(3), optional=True),
            ColumnDefinition('CCF_APID', Number(5), optional=True),
            ColumnDefinition('CCF_NPARS', Number(3), optional=True),
            ColumnDefinition('CCF_PLAN', PlannableColumn(), optional=True, default='N'),
            ColumnDefinition('CCF_EXEC', Boolean(), optional=True, default='Y'),
            ColumnDefinition('CCF_ILSCOPE', CommandInterlockColumn(), optional=True, default='N'),
            ColumnDefinition('CCF_ILSTAGE', CommandVerificationColumn(),
                             optional=True, default='C'),
            ColumnDefinition('CCF_SUBSYS', Number(3), optional=True),
            ColumnDefinition('CCF_HIPRI', Boolean(), optional=True, default='N'),
            ColumnDefinition('CCF_MAPID', Number(2), optional=True),
            ColumnDefinition('CCF_DEFSET', String(8), optional=True),
            ColumnDefinition('CCF_RAPID', Number(5), optional=True),
            ColumnDefinition('CCF_ACK', Number(2), optional=True),
            ColumnDefinition('CCF_SUBSCHEDID', Number(5), optional=True),
            ])


class DSTTable(TableDefinition):
    """The dst.dat table in SCOS

    Command routing"""
    def __init__(self):
        super().__init__('dst', [
            ColumnDefinition('DST_APID', Number(5), optional=False, is_key=True),
            ColumnDefinition('DST_ROUTE', String(30), optional=False),
            ])


class CPCTable(TableDefinition):
    """The cpc.dat table in SCOS

    Command parameters"""
    def __init__(self):
        super().__init__('cpc', [
            ColumnDefinition('CPC_NAME', String(8), optional=False, is_key=True),
            ColumnDefinition('CPC_DESCR', String(24), optional=True),
            ColumnDefinition('CPC_PTC', Number(2), optional=False),
            ColumnDefinition('CPC_PFC', Number(5), optional=False),
            ColumnDefinition('CPC_DISPFMT', CommandInputFormatColumn(),
                             optional=True, default='R'),
            ColumnDefinition('CPC_RADIX', RadixColumn(), optional=True, default='D'),
            ColumnDefinition('CPC_UNIT', String(4), optional=True),
            ColumnDefinition('CPC_CATEG', CommandCategoryColumn(), optional=True, default='N'),
            ColumnDefinition('CPC_PRFREF', String(10), optional=True),
            ColumnDefinition('CPC_CCAREF', String(10), optional=True),
            ColumnDefinition('CPC_PAFREF', String(10), optional=True),
            ColumnDefinition('CPC_INTER', OutputViewColumn(), optional=True, default='R'),
            ColumnDefinition('CPC_DEFVAL', String(248), optional=True),
            ColumnDefinition('CPC_CORR', Boolean(), optional=True, default='Y'),
            ColumnDefinition('CPC_OBTIP', Number(5), optional=True, default='0'),
            ColumnDefinition('CPC_DESCR2', String(256), optional=True, default=''),
            ColumnDefinition('CPC_ENDIAN', EndianessColumn(), optional=True, default='B'),
            ])


class CDFTable(TableDefinition):
    """The cdf.dat table in SCOS

    Command definition
    """
    def __init__(self):
        super().__init__('cdf', [
            ColumnDefinition('CDF_CNAME', String(8), optional=False, is_key=True),
            ColumnDefinition('CDF_ELTYPE', CommandElementTypeColumn(), optional=False),
            ColumnDefinition('CDF_DESCR', String(24), optional=True),
            ColumnDefinition('CDF_ELLEN', Number(4), optional=False),
            ColumnDefinition('CDF_BIT', Number(4), optional=False),
            ColumnDefinition('CDF_GRPSIZE', Number(2), optional=True, default='0'),
            ColumnDefinition('CDF_PNAME', String(8), optional=True),
            ColumnDefinition('CDF_INTER', CommandValueRepresentationColumn(),
                             optional=True, default='R'),
            ColumnDefinition('CDF_VALUE', String(248), optional=True),
            ColumnDefinition('CDF_TMID', String(8), optional=True),
            ])


class PTVTable(TableDefinition):
    """The ptv.dat table in SCOS

    Commands pre-transmission validation"""
    def __init__(self):
        super().__init__('ptv', [
            ColumnDefinition('PTV_CNAME', String(8), optional=False, is_key=True),
            ColumnDefinition('PTV_PARNAM', String(8), optional=False, is_key=True),
            ColumnDefinition('PTV_INTER', OutputViewColumn(), optional=True, default='R'),
            ColumnDefinition('PTV_VAL', String(17), optional=False),
            ])


class CSFTable(TableDefinition):
    """The csf.dat table in SCOS

    Command sequence characteristics"""
    def __init__(self):
        super().__init__('csf', [
            ColumnDefinition('CSF_NAME', String(8), optional=False, is_key=True),
            ColumnDefinition('CSF_DESC', String(24), optional=False),
            ColumnDefinition('CSF_DESC2', String(64), optional=True),
            ColumnDefinition('CSF_IFTT', TimeTagTypeColumn(), optional=True, default='N'),
            ColumnDefinition('CSF_NFPARS', Number(3), optional=True),
            ColumnDefinition('CSF_ELEMS', Number(5), optional=True),
            ColumnDefinition('CSF_CRITICAL', Boolean(), optional=True, default='N'),
            ColumnDefinition('CSF_PLAN', PlannableBySourceColumn(), optional=True, default='N'),
            ColumnDefinition('CSF_EXEC', StandaloneExecutionColumn(), optional=True, default='Y'),
            ColumnDefinition('CSF_SUBSYS', Number(3), optional=True),
            ColumnDefinition('CSF_GENTIME', String(17), optional=True),
            ColumnDefinition('CSF_DOCNAME', String(32), optional=True),
            ColumnDefinition('CSF_ISSUE', String(10), optional=True),
            ColumnDefinition('CSF_DATE', String(17), optional=True),
            ColumnDefinition('CSF_DEFSET', String(8), optional=True),
            ColumnDefinition('CSF_SUBSCHEDID', Number(5), optional=True),
            ])


class CSSTable(TableDefinition):
    """The css.dat table in SCOS

    Command sequence definitions"""
    def __init__(self):
        super().__init__('css', [
            ColumnDefinition('CSS_SQNAME', String(8), optional=False, is_key=True),
            ColumnDefinition('CSS_COMM', String(32), optional=True),
            ColumnDefinition('CSS_ENTRY', Number(5), optional=False),
            ColumnDefinition('CSS_TYPE', SequenceElementTypeColumn(), optional=False),
            ColumnDefinition('CSS_ELEMID', String(8), optional=True),
            ColumnDefinition('CSS_NPARS', Number(3), optional=False),
            ColumnDefinition('CSS_MANDISP', Boolean(), optional=True, default='N'),
            ColumnDefinition('CSS_RELTYPE', ReleaseTimeTypeColumn(), optional=True, default='R'),
            ColumnDefinition('CSS_RELTIME', String(12), optional=True),
            ColumnDefinition('CSS_EXTIME', String(17), optional=True),
            ColumnDefinition('CSS_PREVREL', ReleaseTimeTypeColumn(), optional=True, default='R'),
            ColumnDefinition('CSS_GROUP', RelativePositionColumn(), optional=True),
            ColumnDefinition('CSS_BLOCK', RelativePositionColumn(), optional=True),
            ColumnDefinition('CSS_ILSCOPE', CommandSequenceInterlockColumn(), optional=True),
            ColumnDefinition('CSS_ILSTAGE', CommandVerificationColumn(), optional=True),
            ColumnDefinition('CSS_DYNPTV', Boolean(), optional=True, default='N'),
            ColumnDefinition('CSS_STATPTV', Boolean(), optional=True, default='N'),
            ColumnDefinition('CSS_CEV', Boolean(), optional=True, default='N'),
            ])


class SDFTable(TableDefinition):
    """The sdf.dat table in SCOS

    Command sequence element parameters"""
    def __init__(self):
        super().__init__('sdf', [
            ColumnDefinition('SDF_SQNAME', String(8), optional=False, is_key=True),
            ColumnDefinition('SDF_ENTRY', Number(5), optional=False, is_key=True),
            ColumnDefinition('SDF_ELEMID', String(8), optional=False),
            ColumnDefinition('SDF_POS', Number(4), optional=False, is_key=True),
            ColumnDefinition('SDF_PNAME', String(8), optional=False, is_key=True),
            ColumnDefinition('SDF_FTYPE', ValueEditableColumn(), optional=True, default='E'),
            ColumnDefinition('SDF_VTYPE', ValueSourceColumn(), optional=False),
            ColumnDefinition('SDF_VALUE', String(248), optional=True),
            ColumnDefinition('SDF_VALSET', String(8), optional=True),
            ColumnDefinition('SDF_REPPOS', Number(4), optional=True),
            ])


class CSPTable(TableDefinition):
    """The csp.dat table in SCOS

    Command sequence formal parameters"""
    def __init__(self):
        super().__init__('csp', [
            ColumnDefinition('CSP_SQNAME', String(8), optional=False, is_key=True),
            ColumnDefinition('CSP_FPNAME', String(8), optional=False, is_key=True),
            ColumnDefinition('CSP_FPNUM', Number(8), optional=False),
            ColumnDefinition('CSP_DESCR', String(23), optional=True),
            ColumnDefinition('CSP_PTC', Number(2), optional=False),
            ColumnDefinition('CSP_PFC', Number(5), optional=False),
            ColumnDefinition('CSP_DISPFMT', CommandInputFormatColumn(),
                             optional=True, default='R'),
            ColumnDefinition('CSP_RADIX', RadixColumn(), optional=True, default='D'),
            ColumnDefinition('CSP_TYPE', FormatParameterTypeColumn(), optional=False),
            ColumnDefinition('CSP_VTYPE', OutputViewColumn(), optional=True),
            ColumnDefinition('CSP_DEFVAL', String(248), optional=True),
            ColumnDefinition('CSP_CATEG', CommandCategoryColumn(),
                             optional=True, default='N'),
            ColumnDefinition('CSP_PRFREF', String(10), optional=True),
            ColumnDefinition('CSP_CCAREF', String(10), optional=True),
            ColumnDefinition('CSP_PAFREF', String(10), optional=True),
            ColumnDefinition('CSP_UNIT', String(4), optional=True),
            ])


class PSTTable(TableDefinition):
    """The pst.dat table in SCOS

    Command/Sequence parameter sets"""
    def __init__(self):
        super().__init__('pst', [
            ColumnDefinition('PST_NAME', String(8), optional=False, is_key=True),
            ColumnDefinition('PST_DESCR', String(24), optional=True),
            ])


class PSVTable(TableDefinition):
    """The psv.dat table in SCOS

    Command/sequence parameter value sets"""
    def __init__(self):
        super().__init__('psv', [
            ColumnDefinition('PSV_NAME', String(8), optional=False),
            ColumnDefinition('PSV_PVSID', String(8), optional=False, is_key=True),
            ColumnDefinition('PSV_DESCR', String(24), optional=True),
            ])


class CCATable(TableDefinition):
    """The cca.dat table in SCOS

    Command/Sequence Parameter (De-)calibration curves"""
    def __init__(self):
        super().__init__('cca', [
            ColumnDefinition('CCA_NUMBR', String(10), optional=False, is_key=True),
            ColumnDefinition('CCA_DESCR', String(24), optional=True),
            ColumnDefinition('CCA_ENGFMT', NumericFormatTypeColumn(),
                             optional=True, default='R'),
            ColumnDefinition('CCA_RAWFMT', NumericFormatTypeColumn(),
                             optional=True, default='U'),
            ColumnDefinition('CCA_RADIX', RadixColumn(), optional=True, default='D'),
            ColumnDefinition('CCA_UNIT', String(4), optional=True),
            ColumnDefinition('CCA_NCURVE', Number(3), optional=True),
            ])


class CCSTable(TableDefinition):
    """The ccs.dat table in SCOS

    Command/Sequence Parameter (De-)calibration curve definitions"""
    def __init__(self):
        super().__init__('ccs', [
            ColumnDefinition('CCS_NUMBR', String(10), optional=False, is_key=True),
            ColumnDefinition('CCS_XVALS', String(248), optional=False, is_key=True),
            ColumnDefinition('CCS_YVALS', String(248), optional=False, is_key=True),
            ])


class PAFTable(TableDefinition):
    """The paf.dat table in SCOS

    Command/Sequence Parameter Textual (De-)calibration"""
    def __init__(self):
        super().__init__('paf', [
            ColumnDefinition('PAF_NUMBR', String(10), optional=False, is_key=True),
            ColumnDefinition('PAF_DESCR', String(24), optional=True),
            ColumnDefinition('PAF_RAWFMT', NumericFormatTypeColumn(),
                             optional=True, default='U'),
            ColumnDefinition('PAF_NALIAS', Number(3), optional=True),
            ])


class PASTable(TableDefinition):
    """The pas.dat table in SCOS

    Command/Sequence Parameter Textual (De-)calibration definition"""
    def __init__(self):
        super().__init__('pas', [
            ColumnDefinition('PAS_NUMBR', String(10), optional=False, is_key=True),
            ColumnDefinition('PAS_ALTXT', String(248), optional=False, is_key=True),
            ColumnDefinition('PAS_ALVAL', String(248), optional=False, is_key=True),
            ])


class PRFTable(TableDefinition):
    """The prf.dat table in SCOS

    Command/Sequence Parameter range sets"""
    def __init__(self):
        super().__init__('prf', [
            ColumnDefinition('PRF_NUMBR', String(10), optional=False, is_key=True),
            ColumnDefinition('PRF_DESCR', String(24), optional=True),
            ColumnDefinition('PRF_INTER', OutputViewColumn(), optional=True, default='R'),
            ColumnDefinition('PRF_DSPFMT', CommandInputFormatColumn(),
                             optional=True, default='R'),
            ColumnDefinition('PRF_RADIX', RadixColumn(), optional=True, default='D'),
            ColumnDefinition('PRF_NRANGE', Number(3), optional=True),
            ColumnDefinition('PRF_UNIT', String(4), optional=True),
            ])


class PRVTable(TableDefinition):
    """The prv.dat table in SCOS

    Command/Sequence Parameter range values"""
    def __init__(self):
        super().__init__('prv', [
            ColumnDefinition('PRV_NUMBR', String(10), optional=False, is_key=True),
            ColumnDefinition('PRV_MINVAL', String(248), optional=False, is_key=True),
            ColumnDefinition('PRV_MAXVAL', String(248), optional=True),
            ])


class SCOS:
    """An SCOS-2000 database instance

    This is your main entry point for the use of SCOS-2000 databases.

    To merge this SCOS-2000 database with another database (for example if
    you have an instrument’s database but also a separate database for the
    spacecraft), instatiate both databases and then merge them like this::

        instrument = SCOS('instrumentdb.zip')
        spacecraft = SCOS('spacecraftdb.zip')

        if spacecraft.can_merge(instrument):
            spacecraft.merge(instrument)

    It’s a good idea to run ``can_merge`` first to ensure there are no
    collisions in any of the tables.

    To verify the correctness of the database definition, you can run
    ``check``. It allows you to have a custom set of checkers and also
    a custom reporter, if you don’t want the errors to be written to stderr.

    You can access the various tables (and their rows) like this::

        scos = SCOS('db.zip')
        for row in scos.pcf.rows:
            # do something with each pcfdefinition from the pcf table

    Each table is a ``Table`` instance allowing convenient access to the
    rows with the ``rows`` property and even accessing specific rows by
    their primary key by calling ``.get(...)`` and the primary key values
    as a ``tuple``.

    Each row is a ``TableRow`` instance allowing convenient access to

    - the row’s definition with ``row.column`` (of type ``ColumnDefinition``)
    - the row’s fields by name (``row['PCF_NAME']``) or ``row.get()``
    - the row’s line number via ``row.linenr``
    - all row’s fields by iterating them (``for field in row: ...``)
    """

    class PCFNatureCell(TableCell):
        @property
        def is_synthetic(self):
            return self.table.definition.is_synthetic(self.value)

    class PCFTableInstance(Table):
        def create_cell(self, column, row, field):
            if column.name == 'PCF_NATUR':
                return SCOS.PCFNatureCell(self, column, row, field)
            return super().create_cell(column, row, field)

    class CDFValueCell(TableCell):
        def interprete(self):
            """Return the value interpreted using CDF_PTC and CDF_PFC"""
            if self.value is None:
                return None

            if self.row['CDF_ELTYPE'] == CommandElementType.FIXED:
                value = int(self.value, 16)
                if value < 0:
                    # TODO - extract this into a checker
                    raise RuntimeError("Negative value")
                return value

            if self.row['CDF_INTER'] == CommandValueRepresentation.ENGINEERING:
                return self.value

            cpc = self.table.cpc.get(self.row['CDF_PNAME'])
            assert cpc is not None  # TODO - move this into a checker
            base = cpc['CPC_RADIX']
            if base != Radix.UNDEFINED:
                base = base.radix()
            else:
                base = None

            type_ = ParameterType.resolve((self.row['CDF_PTC'].value,
                                           self.row['CDF_PFC'].value))

            return type_.parse(self.value, base=base)

    class CDFTableInstance(Table):
        def create_cell(self, column, row, field):
            if column.name == 'CDF_VALUE':
                return SCOS.CDFValueCell(self, column, row, field)
            return super().create_cell(column, row, field)

    class CPCDefaultValueCell(TableCell):
        def interprete(self):
            """Return the default value interpreted using CPC_PTC and CPC_PFC"""
            if self.value is None:
                return None

            if self.row['CDF_INTER'] == CommandValueRepresentation.ENGINEERING:
                return self.value

            base = self.row['CPC_RADIX']
            if base != Radix.UNDEFINED:
                base = base.radix
            else:
                base = None

            type_ = ParameterType.resolve((self.row['CPC_PTC'].value,
                                           self.row['CPC_PFC'].value))

            return type_.parse(self.value, base=base)

    class CPCTableInstance(Table):
        def create_cell(self, column, row, field):
            if column.name == 'CPC_DEFVAL':
                return SCOS.CPCDefaultValueCell(self, column, row, field)
            return super().create_cell(column, row, field)

    def __init__(self, path):
        """Instantiate from the SCOS-2000 definition at ``path``.

        ``path`` may either be a folder with the ``.dat`` files or
        a ``.zip`` file containing the ``.dat`` files.
        """
        self.basepath = Path(path)

        self.ziparchive = None
        if self.basepath.is_file() and self.basepath.suffix == '.zip':
            self.ziparchive = ZipFile(self.basepath, 'r')

        self.tables = {
            'vdf': Table(self, VDFTable()).parse(),
            'pcf': SCOS.PCFTableInstance(self, PCFTable()).parse(),
            'cur': Table(self, CURTable()).parse(),
            'caf': Table(self, CAFTable()).parse(),
            'cap': Table(self, CAPTable()).parse().sort(True),
            'txf': Table(self, TXFTable()).parse(),
            'txp': Table(self, TXPTable()).parse().sort(True),
            'mcf': Table(self, MCFTable()).parse(),
            'lgf': Table(self, LGFTable()).parse(),
            'ocf': Table(self, OCFTable()).parse(),
            'ocp': Table(self, OCPTable()).parse(),
            'pid': Table(self, PIDTable()).parse(),
            'pic': Table(self, PICTable()).parse(),
            'tpcf': Table(self, TPCFTable()).parse(),
            'plf': Table(self, PLFTable()).parse(),
            'vpd': Table(self, VPDTable()).parse(),
            # TODO 'grp': Table(self, GRPTable()).parse(),
            # TODO 'grpa': Table(self, GRPATable()).parse(),
            # TODO 'grpk': Table(self, GRPKTable()).parse(),
            # TODO 'dpf': Table(self, DPFTable()).parse(),
            # TODO 'dpc': Table(self, DPCTable()).parse(),
            # TODO 'gpf': Table(self, GPFTable()).parse(),
            # TODO 'gpc': Table(self, GPCTable()).parse(),
            # TODO 'spf': Table(self, SPFTable()).parse(),
            # TODO 'spc': Table(self, SPCTable()).parse(),
            # TODO 'ppf': Table(self, PPFTable()).parse(),
            # TODO 'ppc': Table(self, PPCTable()).parse(),
            'tcp': Table(self, TCPTable()).parse(),
            'pcpc': Table(self, PCPCTable()).parse(),
            'pcdf': Table(self, PCDFTable()).parse(),
            'ccf': Table(self, CCFTable()).parse(),
            'dst': Table(self, DSTTable()).parse(),
            'cpc': SCOS.CPCTableInstance(self, CPCTable()).parse(),
            'cdf': SCOS.CDFTableInstance(self, CDFTable()).parse(),
            'ptv': Table(self, PTVTable()).parse(),
            'csf': Table(self, CSFTable()).parse(),
            'css': Table(self, CSSTable()).parse(),
            'sdf': Table(self, SDFTable()).parse(),
            'csp': Table(self, CSPTable()).parse(),
            # TODO 'cvs': Table(self, CVSTable()).parse(),
            # TODO 'cve': Table(self, CVETable()).parse(),
            # TODO 'cvp': Table(self, CVPTable()).parse(),
            'pst': Table(self, PSTTable()).parse(),
            'psv': Table(self, PSVTable()).parse(),
            # TODO 'cps': Table(self, CPSTable()).parse(),
            # TODO 'pvs': Table(self, PVSTable()).parse(),
            # TODO 'psm': Table(self, PSMTable()).parse(),
            'cca': Table(self, CCATable()).parse(),
            'ccs': Table(self, CCSTable()).parse(),
            'paf': Table(self, PAFTable()).parse(),
            'pas': Table(self, PASTable()).parse(),
            'prf': Table(self, PRFTable()).parse(),
            'prv': Table(self, PRVTable()).parse(),
            }

        self.synthetics = {}
        """The ``synthetics`` dictionary contains the definitions for all
        synthetic parameters. The key is the parameter's name, the value an
        instance of ``SyntheticParameterDefinition``."""
        self.synthetics_in_spid = {}
        """A mapping of SPID to all synthetic parameters that can be evaluated
        in a packet with this SPID (because all input parameters are present in
        that packet according to the PLF table."""
        self.load_synthetics()
        self.parse_synthetics()

        if self.ziparchive is not None:
            self.ziparchive.close()

    def __getattr__(self, name):
        if name in self.tables:
            return self.tables[name]
        raise AttributeError()

    @property
    def version(self):
        """Return the version string of this SCOS2000 instance
        Specifically the VDF release and issue
        """
        return str(self.vdf.rows[0]['VDF_RELEASE'].value) \
             + "." \
             + str(self.vdf.rows[0]['VDF_ISSUE'].value)

    def check(self, checkers=None, reporter=None, pedantic=False):
        """Run all ``checkers`` on this database

        If ``checkers`` is left ``None``, all registered checkers will be run.

        If ``reporter`` is left ``None``, the ``StdErrReporter`` will be used.

        The ``pedantic`` flag can be used to make the checks extra annoying
        (they might check the range for values that are actually ignored in the
        given context).
        """
        if checkers is None:
            checkers = ALL_CHECKERS
        if reporter is None:
            reporter = StdErrReporter()
        for cls in checkers:
            checker = cls(self, reporter, pedantic)
            for table in self.tables.values():
                checker.visit(table)

    def can_merge(self, other):
        """Verify whether the ``other`` SCOS can be merged into this

        This function checks for duplicate/clashing definitions.
        """
        return all(t.can_merge(other.tables[n])
                   for n, t in self.tables.items()) \
               and \
               len(set(self.synthetics.keys()) & set(other.synthetics.keys())) == 0

    def merge(self, other):
        """Merge the SCOS ``other`` into this

        You should check ``can_merge`` before-hand to verify that a
        merge will not cause clashing definitions.
        """
        for name, table in self.tables.items():
            table.merge(other.tables[name])
        self.synthetics.update(other.synthetics)

    def load_synthetics(self):
        """Load all synthetic parameters"""
        self.synthetics = {}

        for row in self.pcf.rows:
            paramdef = None
            if row['PCF_NATUR'].value == ParameterNature.DYNAMIC:
                paramdef = DynamicSyntheticParameter(row)
            elif row['PCF_NATUR'].value == ParameterNature.HARDCODED:
                pass
            elif row['PCF_NATUR'].value == ParameterNature.SYNTHETIC:
                pass
            elif row['PCF_NATUR'].value == ParameterNature.SAVEDSYNTHETIC:
                pass
            else:
                # not a synthetic parameter
                continue

            if paramdef is None:
                raise NotImplementedError(f"PCF_NATUR = {row['PCF_NATUR'].value} "
                                          f"not supported for {row['PCF_NAME'].value}")
            self.synthetics[row['PCF_NAME'].value] = paramdef

    def parse_synthetics(self):
        """Parse and cache the synthetic parameter implementations"""
        for name, definition in self.synthetics.items():
            # find the definition file
            stream = self.open_textfile(name, path=self.basepath / 'synthetic')
            if stream is not None:
                definition.parse(stream)
            else:
                raise SyntheticImplementationMissingError("No implementation found "
                                                          f"for synthetic parameter {name}")

        # build a list of all SPIDs with the parameters that are grouped in them
        spids = {}
        for plf in self.plf.rows:
            spid = plf['PLF_SPID'].value
            if spid not in spids:
                spids[spid] = set()
            spids[spid].add(plf['PLF_NAME'].value)

        self.synthetics_in_spid = {}
        for spid, params in spids.items():
            for name, definition in self.synthetics.items():
                if definition.inputs.issubset(params):
                    if spid not in self.synthetics_in_spid:
                        self.synthetics_in_spid[spid] = set()
                    self.synthetics_in_spid[spid].add(name)

    def open_table(self, tablename):
        """Return a context manager for the file with the given table name

        May return ``None`` if the file for that table does not exist
        """
        tablefilename = tablename + '.dat'
        return self.open_textfile(tablefilename)

    def open_textfile(self, filename, path=None):
        """Return a context manager for reading text from the file with the given name

        May return ``None`` if the file does not exist."""
        if path is None:
            path = self.basepath
        if self.ziparchive is not None:
            fullpath = [fn
                        for fn in self.ziparchive.namelist()
                        if Path(fn).name == filename]
            if len(fullpath) == 0:
                return None
            return ZipArchiveTextReader(self.ziparchive, fullpath[0])

        filepath = path / filename
        if not filepath.is_file():
            return None
        return open(filepath, 'rt', encoding='utf-8', newline='')


class ZipArchiveTextReader:
    """Context manager for reading text files from an open zipfile"""
    def __init__(self, archive, path):
        self.archive = archive
        self.filepath = path
        self.filehandle = None
        self.textstream = None

    def __enter__(self):
        self.filehandle = self.archive.open(self.filepath, 'r')
        self.textstream = io.TextIOWrapper(self.filehandle, encoding='utf-8', newline='')
        return self.textstream

    def __exit__(self, *args, **kwargs):
        if self.textstream is not None:
            self.textstream.close()
        if self.filehandle is not None:
            self.filehandle.close()


class ExtraCAPDefinitions(SCOSTableChecker):
    """Report any CAP definitions that don't have a matching CAF entry"""
    ACCEPTED_TABLES = (CAPTable,)

    def check_row(self, row):
        if (row['CAP_NUMBR'].value,) not in self.scos.caf.index:
            self.error(f"CAP {row['CAP_NUMBR'].value} does "
                       "not have a matching CAF entry.")


class ExtraCAFEntries(SCOSTableChecker):
    """Report any CAF entry that has not the exact right amount of CAP rows"""
    ACCEPTED_TABLES = (CAFTable,)

    def check_row(self, row):
        rows = [r for r in self.scos.cap.rows
                if r['CAP_NUMBR'].value == row['CAF_NUMBR'].value]
        expected = row['CAF_NCURVE'].value
        if len(rows) != expected:
            self.error(f"{row['CAF_NUMBR'].value} should have {expected} CAP entries "
                       f"but has {len(rows)}")


class ExtraCalibrationDefinitions(SCOSTableChecker):
    """Report any unused CAF, TXF, MCF definitions"""
    ACCEPTED_TABLES = (CAFTable, TXFTable, MCFTable)

    def check_row(self, row):
        used = any(r['PCF_CURTX'] == row[0]
                   for r in self.scos.pcf.rows)
        if not used:
            self.error(f"{row[0].value} is not used")


class FieldLengthChecks(SCOSTableChecker):
    """Check that the fields are not longer than accepted by the definition"""

    def check_row(self, row):
        for colidx, cell in enumerate(row):
            rawlen = len(cell.rawvalue)
            coltype = cell.column.type
            allowed = -1

            if isinstance(coltype, String):
                allowed = coltype.maxlen
            elif isinstance(coltype, Number):
                allowed = coltype.digits
            elif isinstance(coltype, Boolean):
                allowed = max(len(coltype.true), len(coltype.false))

            if -1 < allowed < rawlen:
                self.error(f"Cell {colidx+1} ('{cell.rawvalue}') is too long: "
                           f"{rawlen}, max length: {allowed}")

    def accept(self, _):
        return True


class ValidityStateDefinitionChecker(SCOSTableChecker):
    """Check that TM parameters with validity parameters set are okay

    I.e. checking that they have a check value set and that the validy
    parameter type is actually of some integer type.
    """
    ACCEPTED_TABLES = (PCFTable,)

    def check_row(self, row):
        # TODO: check for circular state validation
        # TODO: synthetic parameters may be used for state validation
        validparamname = row['PCF_VALID'].value
        if validparamname is None:
            return
        validparamdef = [other for other in row.table.rows
                         if other['PCF_NAME'] == validparamname]

        if len(validparamdef) == 0:
            self.error(f"{row['PCF_NAME']} uses non-existing {validparamname} as "
                        "state validity parameter.")
            return
        validparamdef = validparamdef[0]

        ptype = ParameterType.resolve(validparamdef)
        if not isinstance(ptype, (SignedIntegerParameter, UnsignedIntegerParameter)):
            self.error(f"{validparamname} is used as a state validity parameter "
                       f"for {row['PCF_NAME']}, but is not signed or unsigned int")
            return


class OnlyAllowedCharactersChecker(SCOSTableChecker):
    """Check that forbidden characters are not used in CHAR columns part of the key"""

    ALLOWED = string.ascii_letters + string.digits + '_+-. '

    def check_row(self, row):
        for colidx, cell in enumerate(row):
            if not isinstance(cell.column.type, String) or not cell.column.is_key:
                continue

            if any(letter not in self.ALLOWED for letter in cell.rawvalue):
                self.error(f"Cell {colidx+1} contains at least one character "
                            "that is not allowed.")
                continue

    def accept(self, _):
        return True


class StaticMonitorValueChecker(SCOSTableChecker):
    """Verify that monitor parameters of nature 'constant' have a PARVAL set"""
    ACCEPTED_TABLES = (PCFTable,)

    def check_row(self, row):
        if row['PCF_NATUR'] != ParameterNature.CONSTANT:
            return

        rawvalue = row['PCF_PARVAL'].rawvalue

        if rawvalue is None or len(rawvalue) == 0:
            self.error(f"PARVAL 'None' (cell 16) not allowed for "
                       f"{row['PCF_NAME']}, because PCF_NATUR is 'C' (constant).")


class BitOffsetChecker(SCOSTableChecker):
    """Verify that bit-offset values are only in the range 0..7"""
    ACCEPTED_TABLES = (PLFTable,)

    def check_row(self, row):
        if not (0 <= int(row['PLF_OFFBI'].value) <= 7):
            self.error(f"{row['PLF_NAME']}: PLF_OFFBI outside 0..7")


class PLFTimeDelayChecker(SCOSTableChecker):
    """Verify that the time delay parameter is in a valid range"""
    ACCEPTED_TABLES = (PLFTable,)

    def check_row(self, row):
        if row['PLF_NBOCC'].value == 1 and not self.pedantic:
            return
        tdocc = row['PLF_TDOCC'].value
        if not (0 < tdocc <= 4080000):
            self.error(f"Invalid TDOCC parameter value {tdocc}")


"""Concrete parameter instances

Convenience classes for handling concrete instances of
telemetry parameters.
"""
import math
from collections import namedtuple

from .parametertypes import (
    ParameterType, SignedIntegerParameter, UnsignedIntegerParameter,
    FloatParameter,
)
from .schema import CalibrationCategory, InterpolationBehaviour, NumericFormatType


class PacketInterface:
    """Abstract packet interface

    When implementing a packet (e.g. PUS packet class) it must satisfy
    this interface to be able to use the SCOS wrapper
    """
    def __getitem__(self, item):
        return self.get(item)

    def __contains__(self, item):
        return self.has(item)

    def get(self, item):
        """Get the parameter ``item``

        ``item`` can be one of these:
        - a number, referring to the position of the parameter in the payload
          of the packet
        - a string, identifying the parameter by name (8 character ``PCF_NAME``)

        If obtaining a parameter by name and there are multiple,
        only the first one is returned.

        Raises ``KeyError`` if the parameter is not there.
        """
        raise NotImplementedError("Must be implemented in a subclass")

    def get_all(self, name):
        """Get all parameters with the identifier ``name``

        Will always return a list, but it might be empty.
        """
        raise NotImplementedError("Must be implemented in a subclass")

    def has(self, item):
        """Whether or not parameter ``item`` is present

        ``item`` can be either a number to select the parameter by position or
        the 8-character parameter identifier.
        """
        raise NotImplementedError("Must be implemented in a subclass")

    def __len__(self):
        """The number of parameters in this packet"""
        raise NotImplementedError("Must be implemented in a subclass")

    def __iter__(self):
        yield from self.iter()

    def iter(self):
        """Iterate through the list of parameters"""
        raise NotImplementedError("Must be implemented in a subclass")


class Parameter:
    """A parameter with its actual values"""
    def __init__(self, packet, pcf):
        self.packet = packet
        self.pcf = pcf
        self.paramtype = ParameterType.resolve(self.pcf)
        self.unit = self.pcf['PCF_UNIT'].value or ''
        self._validity = None
        self._eng_value = None

    def get_hex(self):
        """Return the hex representation of this parameter"""
        raise NotImplementedError()

    def get_raw(self):
        """Return the raw value of this parameter

        The return value must be of the correct `ParameterType`
        """
        raise NotImplementedError()

    def get_eng(self):
        """Return the engineering value of this parameter"""
        if self._eng_value is None:
            self._eng_value = self.calibrate()
        return self._eng_value

    def calibrate(self):
        """Return the calibrated value from the raw value of this parameter"""
        # TODO #4 - evaluate conditional calibration
        calibcat = self.pcf['PCF_CATEG'].value
        calibparamname = self.pcf['PCF_CURTX'].value
        interpolation = self.pcf['PCF_INTER'].value  # TODO - use this!
        if calibparamname is None:
            return str(self.raw)

        calibparam = None
        if calibcat == CalibrationCategory.NUMERICAL:
            # no clear indication, we have to try which calibration is defined

            calibparam = self.scos.mcf.get(calibparamname)
            if calibparam is not None:
                # it is polynomial
                coeff = [float(v.value) for v in calibparam.items[2:]]
                return coeff[0] \
                     + coeff[1]*self.raw \
                     + coeff[2]*(self.raw**2) \
                     + coeff[3]*(self.raw**3) \
                     + coeff[4]*(self.raw**4)

            calibparam = self.scos.lgf.get(calibparamname)
            if calibparam is not None:
                # it is logarithmic
                coeff = [float(v.value) for v in calibparam.items[2:]]
                ln = math.log(self.raw)
                return 1.0/(coeff[0] +
                            coeff[1]*ln +
                            coeff[2]*(ln**2) +
                            coeff[3]*(ln**3) +
                            coeff[4]*(ln**4))

            calibdef = self.scos.caf.get(calibparamname)
            if calibdef is not None:
                # table-based numerical calibration
                return self.calibrate_cafcap(calibdef)

        elif calibcat == CalibrationCategory.STATUS:
            calibparam = self.scos.txf.get(calibparamname)
            rawformat = calibparam['TXF_RAWFMT'].value
            for txp in self.scos.txp.rows:
                if txp['TXP_NUMBR'].value != calibparam['TXF_NUMBR']:
                    continue
                range_from = rawformat.convert(txp['TXP_FROM'].value)
                range_to = rawformat.convert(txp['TXP_TO'].value)
                if range_from <= self.raw <= range_to:
                    return txp['TXP_ALTXT'].value

        elif calibcat == CalibrationCategory.TEXT:
            pass  # TODO #5

        else:
            raise NotImplementedError("Unknown, probably new, calibration "
                                      f"category {calibcat}")

        raise RuntimeError(f"Could not calibrate {self.name} (0x{self.hex}) "
                           f"with {calibparamname}")

    def calibrate_cafcap(self, cafdef):
        """Calibrate the parameter value using the CAF/CAP table"""
        interpolation = cafdef['CAF_INTER'].value
        values = interprete_cap(cafdef)
        lower = values[0]
        upper = values[-1]
        rawvalue = self.raw

        result = None

        # handle invalid extrapolations
        if interpolation == InterpolationBehaviour.INVALID and \
           (rawvalue < lower.x or rawvalue > upper.x):
            return result

        if rawvalue < lower.x:
            factor = (values[1].y - values[0].y)/float(values[1].x - values[0].x)
            result = factor*(rawvalue - values[0].x)
        elif rawvalue > upper.x:
            factor = (values[-2].y - values[-1].y)/float(values[-2].x - values[-1].x)
            result = upper.y + factor*(rawvalue - upper.x)
        else:
            # find the first value that's equal or bigger than rawvalue
            upperidx = 0
            while upperidx < len(values) and \
                  values[upperidx].x < rawvalue:
                upperidx += 1
            lower = values[upperidx-1]
            upper = values[upperidx]

            if rawvalue == lower.x:
                result = lower.y
            elif rawvalue == upper.x:
                result = upper.y
            else:
                factor = (upper.y - lower.y)/float(upper.x - lower.x)
                result = lower.y + factor*(rawvalue - lower.x)

        engfmt = cafdef['CAF_ENGFMT'].value
        if engfmt in (NumericFormatType.SIGNEDINT, NumericFormatType.UNSIGNEDINT):
            return int(result)
        if engfmt == NumericFormatType.REAL:
            return float(result)
        raise TypeError(engfmt)

    @property
    def hex(self):
        """Convenience accessor to ``get_hex``"""
        return self.get_hex()

    @property
    def raw(self):
        """Convenience accessor to ``get_raw``"""
        return self.get_raw()

    @property
    def eng(self):
        """Convenience accessor to ``get_eng``"""
        return self.get_eng()

    @property
    def scos(self):
        """Convenient quick access to the SCOS instance"""
        return self.pcf.scos

    @property
    def name(self):
        """The unique identifier of the parameter"""
        return self.pcf['PCF_NAME'].value

    @property
    def description(self):
        """The description of the parameter"""
        return self.pcf['PCF_DESCR'].value

    @property
    def is_valid(self):
        """Whether or not this parameter is valid"""
        if self._validity is None:
            self._validity = self.calculate_validity()
        return self._validity

    def for_humans(self):
        """Return the human readable version of the parameter

        This will return a tuple of ``(value, unit)``

        ``unit`` may be an empty string if no unit is given or required.

        ``value`` will be the engineering value formatted to the specifications.
        """
        eng = self.eng
        if isinstance(eng, float):
            eng = round(eng, self.pcf['PCF_DECIM'].value)
        return (eng, self.unit)

    def calculate_validity(self):
        """Calculate whether or not this parameter is valid"""
        validparam = self.pcf['PCF_VALID'].value
        if validparam is None:
            return True
        validvalue = self.pcf['PCF_VALPAR'].value

        validparamdef = self.scos.pcf.get(validparam)
        assert validparamdef is not None

        param = None
        if validparamdef['PCF_NATUR'].is_synthetic:
            synthparam = self.scos.synthetics.get(validparam, None)
            if synthparam is None:
                return False
            param = synthparam.evaluate(self.packet)

        if self.packet.has(validparam):
            param = self.packet.get(validparam)

        return param is not None \
           and param.is_valid \
           and param.raw == validvalue

    def __repr__(self):
        return f'<{self.name}: {self.hex}>'

    def __str__(self):
        return self.name


def interprete_cap(caf):
    """Interprete all CAP rows of this CAF table and return the x,y tuple as actual values

    The CAP /Y (raw/eng) values are stored as strings any must be interpreted
    using the CAF_RAWFMT, CAF_RADIX, and CAF_ENGFMT columns. This function
    takes care of it and returns a list of (x, y) tuples (Point instances)
    as their respective python types.
    """
    Point = namedtuple('Point', ('x', 'y'))

    result = []
    radix = caf['CAF_RADIX'].value
    rawfmt = caf['CAF_RAWFMT'].value
    engfmt = caf['CAF_ENGFMT'].value
    caps = [row for row in caf.table.scos.cap.rows
            if row['CAP_NUMBR'].value == caf['CAF_NUMBR'].value]
    for caprow in caps:
        raw = caprow['CAP_XVALS'].value
        if rawfmt == NumericFormatType.UNSIGNEDINT:
            raw = radix.convert(raw)
        else:
            raw = rawfmt.convert(raw)
        eng = engfmt.convert(caprow['CAP_YVALS'].value)
        result.append(Point(raw, eng))
    return result



"""Synthetic parameter handling classes and functions"""

from .concrete import Parameter


class SyntheticParameterDefinition:
    """The definition of a synthetic parameter"""
    def __init__(self, pcf):
        self.inputs = set()
        """The set of parameters (by ID) that are inputs/dependencies to this
        synthetic parameter.

        This set will be used to determine in what packets this parameter will
        magically appear."""
        self.pcf = pcf

    def evaluate(self, packet):
        """Evaluate the definition of this parameter in the context of this ``packet``

        ``packet`` should be derived from ``PacketInterface`` (or otherwise
        satisfy the interface).

        Returns a ``SyntheticParameterInstance`` containing the respective ``raw``
        and ``eng`` values.
        """
        raise NotImplementedError()

    def parse(self, stream):
        """Parse the parameter definition from ``stream``"""
        raise NotImplementedError()


class DynamicSyntheticParameter(SyntheticParameterDefinition):
    """A dynamic synthetic parameter (PCF_NATUR = D)"""
    # TODO #7 - that's a big project on its own right here:
    # this should use EGOS-MCS-S2K-SUM-0019 to parse the
    # OL definition of such a synthetic parameter

    def evaluate(self, packet):
        # for now all we do is to provide an instance that's just plain invalid
        return InvalidSyntheticParameterInstance(packet, self.pcf)

    def parse(self, stream):
        pass


class SyntheticParameterInstance(Parameter):
    """A synthetic parameter instance with a fixed value"""
    def __init__(self, rawvalue, packet, pcf):
        super().__init__(packet, pcf)
        self._raw_value = rawvalue

    def get_raw(self):
        return self._raw_value


class InvalidSyntheticParameterInstance(SyntheticParameterInstance):
    """A generic invalid parameter instance"""
    def __init__(self, packet, pcf):
        super().__init__(0, packet, pcf)

    def calculate_validity(self):
        return False



"""Various shared functions and constants"""
import math

UINT_BYTE_SIZES = {1: 'B',
                   2: 'H',
                   4: 'I',
                   8: 'Q'}
INT_BYTE_SIZES = {1: 'b',
                  2: 'h',
                  4: 'i',
                  8: 'q'}


def unpack_uint_format(bitwidth):
    """Return the unpack format qualifier and expected byte size
       for unsigned ints of this bitwidth"""
    assert 1 <= bitwidth <= 32
    bytewidth = math.ceil(bitwidth/8.)
    while bytewidth not in UINT_BYTE_SIZES:
        bytewidth += 1
    return UINT_BYTE_SIZES[bytewidth], bytewidth


def unpack_int_format(bitwidth):
    """Return the unpack format qualifier and expected byte size
       for signed ints of this bitwidth"""
    assert 1 <= bitwidth <= 32
    bytewidth = math.ceil(bitwidth/8.)
    while bytewidth not in INT_BYTE_SIZES:
        bytewidth += 1
    return INT_BYTE_SIZES[bytewidth], bytewidth


class Error(RuntimeError):
    """Base error for all pyscos2000 errors"""



"""SCOS table schema definition"""
import enum


class IndicatorType(enum.Enum):
    """Event indicator types"""
    UNKNOWN = None
    """Undefined"""

    NOEVENT = 'N'
    """No event should be triggered"""

    INFO = 'I'
    """An event of 'info' severity should be triggered"""

    WARNING = 'W'
    """An event of 'warning' severity should be triggered"""

    ALARM = 'A'
    """An event of 'alarm' severity should be triggered"""


class CalibrationCategory(enum.Enum):
    NUMERICAL = 'N'
    STATUS = 'S'
    TEXT = 'T'


class ParameterNature(enum.Enum):
    RAW = 'R'
    DYNAMIC = 'D'
    SYNTHETIC = 'P'
    HARDCODED = 'H'
    SAVEDSYNTHETIC = 'S'
    CONSTANT = 'C'


class InterpolationBehaviour(enum.Enum):
    UNDEFINED = None
    EXTRAPOLATE = 'P'
    INVALID = 'F'


class OutputView(enum.Enum):
    ENGINEERING = 'E'
    RAW = 'R'


class Endianess(enum.Enum):
    BIG_ENDIAN = 'B'
    LITTLE_ENDIAN = 'L'


class Justification(enum.Enum):
    LEFT = 'L'
    RIGHT = 'R'
    CENTER = 'C'


class IntegerFormatType(enum.Enum):
    SIGNEDINT = 'I'
    UNSIGNEDINT = 'U'


class NumericFormatType(enum.Enum):
    SIGNEDINT = 'I'
    UNSIGNEDINT = 'U'
    REAL = 'R'

    def convert(self, text):
        """Convert the given text into this format's type

        E.g. if ``self == NumericFormatType.REAL`` convert will return a ``float``
        """
        if self == NumericFormatType.REAL:
            return float(text)
        if self in [NumericFormatType.SIGNEDINT, NumericFormatType.UNSIGNEDINT]:
            return int(text)
        return ValueError()


class Radix(enum.Enum):
    """The radix of an unsigned integer value's text representation"""
    UNDEFINED = None
    DECIMAL = 'D'
    HEXADECIMAL = 'H'
    OCTAL = 'O'

    def radix(self):
        """Return the radix value"""
        if self == Radix.DECIMAL:
            return 10
        if self == Radix.HEXADECIMAL:
            return 16
        if self == Radix.OCTAL:
            return 8
        raise ValueError(f"Unspecified radix {self}")

    def convert(self, text):
        """Convert the given text to an unsigned int given this radix"""
        return int(text, self.radix())


class LimitInterpretation(enum.Enum):
    """What value to use to interprete limit values"""

    UNCALIBRATED = 'U'
    """Use RAW values"""
    CALIBRATED = 'C'
    """Use ENG values"""


class LimitTypeInterpretation(enum.Enum):
    """What type the value is that is used to interprete limit values"""

    REAL = 'R'
    """Floating point"""

    TEXT = 'A'
    """ASCII text"""

    INTEGER = 'I'
    """Signed or unsigned integer values"""


class MonitoringCheckFlag(enum.Enum):
    """Of what type a monitoring check is"""

    SOFT_LIMIT = 'S'
    HARD_LIMIT = 'H'
    DELTA = 'D'
    CONSISTENCY = 'C'
    EVENT_ONLY = 'E'


class PacketHeaderElementType(enum.Enum):
    FIXED = 'F'
    APID = 'A'
    SERVICETYPE = 'T'
    SERVICESUBTYPE = 'S'
    ACK = 'K'
    OTHER = 'P'
    UNDEF1 = 'O'  # found in the wild


class Plannable(enum.Enum):
    ALL = 'A'
    FLIGHTDYNAMICS = 'F'
    SOC = 'S'
    NOT = 'N'


class CommandInterlock(enum.Enum):
    GLOBAL = 'G'
    LOCAL = 'L'
    SUBSYSTEM = 'S'
    BELONG = 'B'
    NONE = 'N'


class CommandVerification(enum.Enum):
    RECEPTION = 'R'
    ACCEPTANCE = 'R'
    UPLINK = 'U'
    ONBOARDRECEPTION = 'O'
    ONBOARDACCEPTANCE = 'O'
    APPACCEPTANCE = 'A'
    COMPLETION = 'C'


class CommandInputFormat(enum.Enum):
    ASCII = 'A'
    SIGNEDINT = 'I'
    UNSIGNEDINT = 'U'
    REAL = 'R'
    ABSTIME = 'T'
    RELTIME = 'D'


class CommandCategory(enum.Enum):
    NUMERIC = 'C'
    TEXTUAL = 'T'
    BOTH = 'B'
    COMMANDID = 'A'
    PARAMETERID = 'P'
    NONE = 'N'


class DisplayFormat(enum.Enum):
    BINARY = 'B'
    OCTAL = 'O'
    DECIMAL = 'D'
    HEXADECIMAL = 'H'
    NORMAL = 'N'


class TimeTagType(enum.Enum):
    YES = 'Y'
    NONE = 'N'
    ONBOARD = 'B'


class PlannableBySource(enum.Enum):
    ALL = 'A'
    FLIGHT_DYNAMICS = 'F'
    SOC = 'S'
    NOT_PLANNABLE = 'N'


class StandaloneExecution(enum.Enum):
    ALLOWED = 'Y'
    NOT_ALLOWED = 'N'


class SequenceElementType(enum.Enum):
    COMMAND = 'C'
    SEQUENCE = 'S'
    COMMENT = 'T'
    FPARAM_COMMAND = 'F'
    FPARAM_SEQUENCE = 'P'


class ReleaseTimeType(enum.Enum):
    RELATIVE_PREVIOUS = 'R'
    RELATIVE_START = 'A'


class RelativePosition(enum.Enum):
    START = 'S'
    MIDDLE = 'M'
    END = 'E'


class CommandSequenceInterlock(enum.Enum):
    GLOBAL = 'G'
    LOCAL = 'L'
    SUBSYSTEM = 'S'
    BELONG = 'B'
    LOCALFAILURE = 'F'
    SUBSYSTEMFAILURE = 'T'
    NONE = 'N'


class CommandValueRepresentation(enum.Enum):
    RAW = 'R'
    ENGINEERING = 'E'
    DEFAULT = 'D'
    DYNAMIC_DEFAULT = 'T'


class ValueEditable(enum.Enum):
    """Whether the parameter value of a sequence element parameter is modifiable
    after after loading it onto the stack

    SDF_FTYPE field of SDF table
    """
    EDITABLE = 'E'
    FIXED = 'F'

    def __bool__(self):
        return self == ValueEditable.EDITABLE


class ValueSource(enum.Enum):
    """Command sequence element parameter source

    SDF_VTYPE field of SDF table
    """

    RAW = 'R'
    """Value is taken from SDF_VALUE and interpreted as raw"""

    ENG = 'E'
    """Value is taken from SDF_VALUE and interpreted as engineering"""

    FORMAL = 'F'
    ELEMENT = 'P'
    SEQUENCE = 'S'

    DEFAULT = 'D'
    """For command parameters, the value should be taken from the default value"""


class FormatParameterType(enum.Enum):
    """Type of a command sequence formal parameter

    CSP_TYPE field of CSP table"""
    COMMAND = 'C'
    SEQUENCE = 'S'
    PARAMETER = 'P'


class CommandElementType(enum.Enum):
    """Command element type

    CDF_ELTYPE field of CDF table
    """
    AREA = 'A'
    FIXED = 'F'
    EDITABLE = 'E'


class TableColumnType:
    """The type of a column in an SCOS table definition"""

    def parse(self, text):
        """Parse the raw ``text`` and return the converted value"""
        if text is None:
            return None
        return self.do_parse(text)

    def do_parse(self, text):
        """Delegated parsing of the raw `text`"""
        raise NotImplementedError()


class Number(TableColumnType):
    """A numeric SCOS table column type"""
    def __init__(self, digits):
        super().__init__()
        self.digits = digits

    def do_parse(self, text):
        if text.startswith('0x'):
            return int(text[2:], 16)
        if text.startswith('0') and len(text) > 1:
            return int(text[1:], 8)
        if '.' in text or 'e' in text.lower():
            return float(text)
        return int(text)


class Boolean(TableColumnType):
    """A boolean SCOS table column type"""
    def __init__(self, true='Y', false='N'):
        super().__init__()
        self.true = true
        self.false = false

    def do_parse(self, text):
        return text == self.true


class String(TableColumnType):
    """A string of characters SCOS table column type"""
    def __init__(self, maxlen):
        super().__init__()
        self.maxlen = maxlen

    def do_parse(self, text):
        return text


class EnumColumnType(TableColumnType):
    """Generic definition of a column mapping to an enum type"""
    ENUM = None
    def do_parse(self, text):
        assert self.ENUM is not None
        return self.ENUM(text)


class EventPacketIndicator(EnumColumnType):
    """Behaviour description type for SCOS table columns"""
    ENUM = IndicatorType


class CalibrationCategoryColumn(EnumColumnType):
    """Calibration category type for SCOS table columns"""
    ENUM = CalibrationCategory


class ParameterNatureColumn(EnumColumnType):
    """Nature of a parameter as type for SCOS table columns"""
    ENUM = ParameterNature


class InterpolationColumn(EnumColumnType):
    """Interpolation behaviour definition for SCOS table columns"""
    ENUM = InterpolationBehaviour


class OutputViewColumn(EnumColumnType):
    """Parameter output view SCOS table column"""
    ENUM = OutputView


class EndianessColumn(EnumColumnType):
    """Endianess definition for an SCOS column"""
    ENUM = Endianess


class JustifyColumn(EnumColumnType):
    """Justification definition in ``vpd.dat``"""
    ENUM = Justification


class IntegerFormatTypeColumn(EnumColumnType):
    """Column identifying an integer format type"""
    ENUM = IntegerFormatType


class NumericFormatTypeColumn(EnumColumnType):
    """Column identifying a format type"""
    ENUM = NumericFormatType


class RadixColumn(EnumColumnType):
    """Column specifying the radix for unsigned integer display"""
    ENUM = Radix


class LimitInterpretationColumn(EnumColumnType):
    """Column with the flag for limit value interpretation"""
    ENUM = LimitInterpretation


class LimitTypeInterpretationColumn(EnumColumnType):
    """Column with the type for limit value interpretation"""
    ENUM = LimitTypeInterpretation


class MonitoringCheckFlagColumn(EnumColumnType):
    """Column for the type of monitoring checks"""
    ENUM = MonitoringCheckFlag


class PacketHeaderElementTypeColumn(EnumColumnType):
    """Column with packet header element type for PCDF table"""
    ENUM = PacketHeaderElementType


class PlannableColumn(EnumColumnType):
    """Column with Plannable type for command characteristics"""
    ENUM = Plannable


class CommandInterlockColumn(EnumColumnType):
    """Column with command interlock behaviour"""
    ENUM = CommandInterlock


class CommandVerificationColumn(EnumColumnType):
    """Column for the command verification type"""
    ENUM = CommandVerification


class CommandInputFormatColumn(EnumColumnType):
    """Column for CommandInputFormat in the command paramater table"""
    ENUM = CommandInputFormat


class CommandCategoryColumn(EnumColumnType):
    """Column for command category"""
    ENUM = CommandCategory


class DisplayFormatColumn(EnumColumnType):
    """Column for Display Format settings"""
    ENUM = DisplayFormat


class TimeTagTypeColumn(EnumColumnType):
    ENUM = TimeTagType


class PlannableBySourceColumn(EnumColumnType):
    ENUM = PlannableBySource


class StandaloneExecutionColumn(EnumColumnType):
    ENUM = StandaloneExecution


class SequenceElementTypeColumn(EnumColumnType):
    ENUM = SequenceElementType


class ReleaseTimeTypeColumn(EnumColumnType):
    ENUM = ReleaseTimeType


class RelativePositionColumn(EnumColumnType):
    ENUM = RelativePosition


class CommandSequenceInterlockColumn(EnumColumnType):
    ENUM = CommandSequenceInterlock


class ValueEditableColumn(EnumColumnType):
    ENUM = ValueEditable


class ValueSourceColumn(EnumColumnType):
    ENUM = ValueSource


class FormatParameterTypeColumn(EnumColumnType):
    ENUM = FormatParameterType


class CommandElementTypeColumn(EnumColumnType):
    ENUM = CommandElementType


class CommandValueRepresentationColumn(EnumColumnType):
    ENUM = CommandValueRepresentation


class ColumnDefinition:
    """A column in an SCOS table definition"""
    def __init__(self,
                 name,
                 type_,
                 description="",
                 optional=False,
                 default=None,
                 is_key=False):
        self.name = name
        self.type = type_
        self.description = description
        self.optional = optional
        self.default = default
        self.is_key = is_key


class TableDefinition:
    """A generic definition of an SCOS table"""
    def __init__(self, name, columns):
        self.name = name
        self.columns = columns
        self.primary_key = [cidx
                            for cidx, column in enumerate(self.columns)
                            if column.is_key]
        self.columnnames = [c.name for c in self.columns]
        self.filename = self.name + '.dat'

    @property
    def has_primary_key(self):
        """Whether or not a primary key is defined"""
        return len(self.primary_key) > 0


"""Parameter Types as defined by SCOS-2000

These are used for both unpacking binary data as well as parsing (default)
values from an instance of SCOS-2000 (e.g. ``CPC_VALUE``).

Suppose you have a PCF definition (with PTC and PFC values) and want to use
that to unpack a byte block (``blob``). In that case you would use the provided
``ParameterType`` class like this::

    ptype = ParameterType.resolve(my_pcf_row)
    value = ptype.unpack(blob)

"""
import struct

from . import shared


_parametercache = {}


class NotEnoughDataToUnpack(shared.Error):
    """Raised when attempting to unpack data from a blob
       that's not big enough to contain that data"""
    def __init__(self, expected, available):
        super().__init__(f"Not enough data to unpack {expected} bytes: "
                         f"{available} bytes available")


class ParameterType:
    """Generic parameter type"""

    @classmethod
    def resolve(cls, pcf):
        """Return an instance of the ParameterType for this parameter characteristics

        May raise ``KeyError`` for invalid ptc/pfc combinations.
        """
        if isinstance(pcf, tuple):
            ptc, pfc = pcf
        else:
            ptc = pcf['PCF_PTC'].value
            pfc = pcf['PCF_PFC'].value

        paramtype = _parametercache.get((ptc, pfc), None)

        if paramtype is not None:
            return paramtype

        if ptc == 1 and pfc == 0:
            paramtype = UnsignedIntegerParameter(1)

        elif ptc == 2:
            # enumerations are special and have their own cache
            paramtype = EnumParameter(pfc)

        elif ptc == 3:
            bitwidth = 0

            if 0 <= pfc <= 12:
                bitwidth = 4 + pfc
            elif pfc == 13:
                bitwidth = 24
            elif pfc == 14:
                bitwidth = 32
            elif pfc == 15:
                raise NotImplementedError("unsigned int with 48 bit is not supported")
            elif pfc == 16:
                raise NotImplementedError("unsigned int with 64 bit is not supported")

            if bitwidth > 0:
                paramtype = UnsignedIntegerParameter(bitwidth)

        elif ptc == 4:
            bitwidth = 0

            if 0 <= pfc <= 12:
                bitwidth = 4 + pfc
            elif pfc == 13:
                bitwidth = 24
            elif pfc == 14:
                bitwidth = 32
            elif pfc == 15:
                raise NotImplementedError("signed int with 48 bit is not supported")
            elif pfc == 16:
                raise NotImplementedError("signed int with 64 bit is not supported")

            if bitwidth > 0:
                paramtype = SignedIntegerParameter(bitwidth)

        elif ptc == 5:
            if pfc == 1:
                # IEEE simple prec. float, 32 bit
                paramtype = FloatParameter()
            elif pfc == 2:
                # IEEE double prec. float, 64 bit
                paramtype = DoubleParameter()
            elif pfc == 3:
                # MIL simple prec. float, 32 bit
                raise NotImplementedError()
            elif pfc == 4:
                # MIL double prec. float, 48 bit
                raise NotImplementedError()

        elif ptc == 6:
            if pfc == 0:
                raise NotImplementedError("variable bitlength not supported")

            elif 1 <= pfc <= 32:
                paramtype = UnsignedIntegerParameter(pfc)

        elif ptc == 7:
            if pfc == 0:
                raise NotImplementedError("Variable length octet-string not implemented")

            else:
                paramtype = OctetStringParameter(pfc)

        elif ptc == 8:
            if pfc == 0:
                raise NotImplementedError("Variable length text strings not implemented")

            else:
                paramtype = TextStringParameter(pfc)

        elif ptc == 9:
            if pfc == 1:
                paramtype = AbsoluteCDSTimeFormat(6)
            elif pfc == 2:
                paramtype = AbsoluteCDSTimeFormat(8)
            elif pfc == 3:
                paramtype = AbsoluteCUCTimeFormat(1, 0)
            elif pfc == 4:
                paramtype = AbsoluteCUCTimeFormat(1, 1)
            elif pfc == 5:
                paramtype = AbsoluteCUCTimeFormat(1, 2)
            elif pfc == 6:
                paramtype = AbsoluteCUCTimeFormat(1, 3)
            elif pfc == 7:
                paramtype = AbsoluteCUCTimeFormat(2, 0)
            elif pfc == 8:
                paramtype = AbsoluteCUCTimeFormat(2, 1)
            elif pfc == 9:
                paramtype = AbsoluteCUCTimeFormat(2, 2)
            elif pfc == 10:
                paramtype = AbsoluteCUCTimeFormat(2, 3)
            elif pfc == 11:
                paramtype = AbsoluteCUCTimeFormat(3, 0)
            elif pfc == 12:
                paramtype = AbsoluteCUCTimeFormat(3, 1)
            elif pfc == 13:
                paramtype = AbsoluteCUCTimeFormat(3, 2)
            elif pfc == 14:
                paramtype = AbsoluteCUCTimeFormat(3, 3)
            elif pfc == 15:
                paramtype = AbsoluteCUCTimeFormat(4, 0)
            elif pfc == 16:
                paramtype = AbsoluteCUCTimeFormat(4, 1)
            elif pfc == 17:
                paramtype = AbsoluteCUCTimeFormat(4, 2)
            elif pfc == 18:
                paramtype = AbsoluteCUCTimeFormat(4, 3)
            elif pfc == 30:
                paramtype = AbsoluteUnixTimeFormat()

        if paramtype is not None:
            _parametercache[(ptc, pfc)] = paramtype
            return paramtype

        raise KeyError(f"Invalid PTC/PFC combination {ptc}/{pfc}")

    def __init__(self, bitwidth):
        self.bitwidth = bitwidth

    def parse(self, text, base=10):
        """Parse the given text in the given base (if applicable)"""
        raise NotImplementedError("Must be implemented in subclasses")

    def unpack(self, blob):
        """Unpack the bytes at `blob` and return the respective value"""
        raise NotImplementedError("Must be implemented in subclasses")


class UnsignedIntegerParameter(ParameterType):
    """An unsigned integer parameter type of variable bitwidth"""
    def unpack(self, blob):
        mask = 2**self.bitwidth - 1
        fmt, width = shared.unpack_uint_format(self.bitwidth)
        while len(blob) < width:
            blob = b'\x00' + blob
        return struct.unpack_from('>'+fmt, blob)[0] & mask

    def parse(self, text, base=10):
        value = int(text, base)

        if value < 0:
            raise ValueError("Negative value")

        return value


class EnumParameter(UnsignedIntegerParameter):
    """Enums are unsigned int"""


class SignedIntegerParameter(ParameterType):
    """A signed integer parameter type of variable bitwidth"""
    def unpack(self, blob):
        mask = 2**self.bitwidth - 1
        fmt, width = shared.unpack_int_format(self.bitwidth)
        while len(blob) < width:
            blob = b'\x00' + blob
        return struct.unpack_from('>'+fmt, blob)[0] & mask

    def parse(self, text, base=10):
        return int(text, base)


class FloatParameter(ParameterType):
    """A simple precision floating point type"""
    def __init__(self):
        super().__init__(32)

    def unpack(self, blob):
        if len(blob) < 4:
            raise NotEnoughDataToUnpack(4, len(blob))
        return struct.unpack_from('>f', blob)[0]

    def parse(self, text, base=None):
        return float(text)


class DoubleParameter(ParameterType):
    """A double precision floating point type"""
    def __init__(self):
        super().__init__(64)

    def unpack(self, blob):
        if len(blob) < 8:
            raise NotEnoughDataToUnpack(8, len(blob))
        return struct.unpack_from('>d', blob)[0]

    def parse(self, text, base=None):
        return float(text)


class OctetStringParameter(ParameterType):
    """A fixed length octet string"""
    def unpack(self, blob):
        # TODO - how is PTC = 7, PFC > 0 interpreted?
        raise NotImplementedError()

    def parse(self, text, base=None):
        raise NotImplementedError()


class TextStringParameter(ParameterType):
    """A fixed length text string"""
    def __init__(self, bitwidth):
        super().__init__(bitwidth*8)

    def unpack(self, blob):
        if len(blob)*8 < self.bitwidth:
            raise NotEnoughDataToUnpack(self.bitwidth, len(blob))
        return str(blob[:self.bitwidth], encoding="ascii", errors="ignore")

    def parse(self, text, base=None):
        return text


class AbsoluteCUCTime:
    """A CCSDS Unsegmented Time Code Timestamp"""
    def __init__(self, coarse, fine):
        self.coarsebytes = coarse
        self.finebytes = fine

    def parse(self, blob):
        # TODO
        return self


class AbsoluteCUCTimeFormat(ParameterType):
    """CCSDS Unsegmented Time Code"""
    def __init__(self, coarsetimebytes, finetimebytes):
        super().__init__((coarsetimebytes + finetimebytes)*8)
        self.coarse = coarsetimebytes
        self.fine = finetimebytes

    def unpack(self, blob):
        if len(blob) < self.coarse + self.fine:
            raise NotEnoughDataToUnpack(self.bitwidth, len(blob))
        return AbsoluteCUCTime(self.coarse, self.fine).parse(blob)

    def parse(self, text, base=None):
        raise NotImplementedError("No textual representation possible for CUC time stamps")


class AbsoluteCDSTimeFormat(ParameterType):
    """CDS time stamp without p-field"""
    def __init__(self, octets):
        super().__init__(octets*8)
        self.octets = octets

    def unpack(self, blob):
        if len(blob) < self.octets:
            raise NotEnoughDataToUnpack(self.bitwidth, len(blob))
        raise NotImplementedError()


class AbsoluteUnixTimeFormat(ParameterType):
    """Absolute timestamp in 8 byte unix time format"""
    def __init__(self):
        super().__init__(64)

    def unpack(self, blob):
        if len(blob) < 8:
            raise NotEnoughDataToUnpack(64, len(blob))
        raise NotImplementedError()


assert 6 == ParameterType.resolve((3, 0)).unpack(b'\xa6')
assert 38 == ParameterType.resolve((3, 2)).unpack(b'\xa6')
assert 0x63c == ParameterType.resolve((3, 8)).unpack(b'\xa6\x3c')


"""SCOS Table instance classes"""
import csv


class TableCell:
    """A cell in an SCOS table"""
    def __init__(self, table, column, row, rawvalue):
        self.table = table
        self.column = column
        self.row = row
        assert column in self.table.definition.columns

        self.columnidx = self.table.definition.columns.index(column)
        self.rawvalue = rawvalue
        if len(rawvalue) > 0:
            self._value = column.type.parse(rawvalue)
        else:
            self._value = column.type.parse(column.default)

    @property
    def value(self):
        """The parsed value of the cell"""
        return self._value

    @property
    def is_key(self):
        """Whether or not this cell is part of the row's primary key"""
        return self.column.is_key

    @property
    def type(self):
        """The column type of this cell"""
        return self.column.type

    def __len__(self):
        if len(self.rawvalue) == 0:
            return 0

        return len(self._value)

    def __eq__(self, other):
        if isinstance(other, TableCell):
            return self._value == other._value
        if isinstance(other, type(self._value)):
            return self._value == other
        raise ValueError(f"{type(other).__name__} not comparable to {type(self._value)}")

    def __lt__(self, other):
        if self._value is None:
            return True
        if isinstance(other, TableCell):
            return self._value < other._value
        if isinstance(other, type(self._value)):
            return self._value < other
        raise ValueError(f"{type(other).__name__} not comparable to {type(self._value)}")

    def __str__(self):
        return f'{self.column.name}: {self.value} ({self.rawvalue})'

    def __repr__(self):
        return str(self)


class TableRow:
    """A row in an SCOSTable"""

    def __init__(self, table, linenr):
        self.table = table
        self.items = []
        self.linenr = linenr

    def __len__(self):
        return len(self.items)

    def __getitem__(self, item):
        return self.get(item)

    def __iter__(self):
        yield from self.items

    @property
    def scos(self):
        """Convenience access to the SCOS instance"""
        return self.table.scos

    def key(self):
        """Return the primary key tuple identifying this row

        Will raise a ``RuntimeError`` if no primary key columns are defined
        """
        return tuple(cell.value
                     for cell in self.items
                     if cell.is_key)

    def get(self, item):
        """Obtain row item by index or column name"""
        if isinstance(item, int):
            return self.items[item]
        if isinstance(item, str):
            return self.items[self.table.definition.columnnames.index(item)]
        raise TypeError()

    def append(self, item):
        """Append this item to the end of the row"""
        self.items.append(item)

    def __lt__(self, other):
        return self.key() < other.key()

    def __str__(self):
        return str(self.items)

    def __repr__(self):
        return repr(self.items)


class BasicTable:
    """An instance of an SCOS table

    The table represents a parsed SCOS table with its rows
    available as the property ``rows``.
    The table definition is available as ``definition``, but you should
    never have to access it.
    Instead each row consists of a number of ``SCOSTableCell`` wich each contain
    the value in that column of that row, and have a reference to their
    column definition.
    """

    ROW_TYPE = TableRow
    CELL_TYPE = TableCell

    def __init__(self, scos, definition):
        self.scos = scos
        self.definition = definition
        self.rows = []

    @property
    def name(self):
        return self.definition.name

    def get(self, key):
        """Return the row of this key using the first few columns

        Depending on how many columns your ``key`` has, that many are used to
        identify a row.
        This is a close approximation to the primary key behaviour for tables
        that don't have a primary key.
        It will return the first found row. Which may or may not be the only
        row matching the ``key``.

        But all in all you're probably better off using ``find``.

        Returns ``None`` if entry is not found
        """
        if isinstance(key, list):
            key = tuple(key)
        if not isinstance(key, tuple):
            key = (key,)
        key = tuple(k.value if isinstance(k, self.CELL_TYPE) else k
                    for k in key)
        for row in self.rows:
            if all(row.items[i].value == key[i]
               for i in range(len(key))):
                return row
        return None

    def find(self, values):
        """Return a list of all matching rows given these ``values``

        - ``values`` can be a tuple or list to match against the columns in order,
        - ``values`` can also be a dictionary to match the respective named column
          against the given value

        In both use-cases the values may also be of type TableCell.
        """
        if isinstance(values, dict):
            for row in self.rows:
                matches = all(sval.value == row[name].value
                              if isinstance(sval, self.CELL_TYPE)
                              else sval == row[name].value
                              for name, sval in values.items())
                if matches:
                    yield row

        if isinstance(values, (tuple, list)):
            values = [v.value
                      if isinstance(v, self.CELL_TYPE)
                      else v
                      for v in values]
            for row in self.rows:
                matches = all(value == row[idx].value
                              for idx, value in enumerate(values))
                if matches:
                    yield row

    def parse(self):
        """Parse the table name"""
        context = self.scos.open_table(self.name)
        if context is not None:
            with context as tabbed:
                reader = csv.reader(tabbed, dialect=CSVDialect)
                self._parse_from_reader(reader)
        return self

    def _parse_from_reader(self, reader):
        """Parse the rows from the given csv.reader instance"""
        for linenr, row in enumerate(reader):
            tablerow = self.create_row(linenr+1)
            self.rows.append(tablerow)
            for columnidx, field in enumerate(row):
                column = self.definition.columns[columnidx]
                tablerow.append(self.create_cell(column, tablerow, field))

    def create_cell(self, column, row, field):
        """Return a new instance of ``CELL_TYPE`` for this column"""
        return self.CELL_TYPE(self, column, row, field)

    def create_row(self, linenr):
        """Return a new instance of ``ROW_TYPE`` for this line"""
        return self.ROW_TYPE(self, linenr)

    def can_merge(self, other):
        """Whether or not the ``other`` table's content can be merged

        Conditions are:

        - no clashing entries in the primary keys
        - same table name
        """
        return self.definition.filename == other.definition.filename

    def merge(self, other):
        """Merge the ``other`` table's content into this table"""
        self.rows += other.rows
        self.rows.sort()

    def __iter__(self):
        yield from self.rows

    def sort(self, full=False):
        """Sort the rows of this table

        By default only the key column(s) are used for sorting.

        If there are no key columns or ``full`` is ``True``, all columns
        will be used for sorting.

        The rows are sorted in place, but the function returns this instance
        of the table, for convenience.
        """
        sortcolumns = self.definition.primary_key[:]
        if len(sortcolumns) == 0 or full:
            sortcolumns = range(len(self.definition.columns))

        self.rows.sort(key=lambda r: [r.items[i] for i in sortcolumns])

        return self


class Table(BasicTable):
    """The same as ``pyscos2000.instance.BasicTable``, but with an index

    There is also an index over the rows based on the definition's primary
    key. You can use the ``get`` method to obtain rows by their index.
    """
    def __init__(self, scos, definition):
        super().__init__(scos, definition)
        self.index = {}

    def get(self, key):
        """Return the row of this key using the index

        Usually ``key`` is of type ``tuple``, but for the lazy developer there
        are a few conveniences built in:

        - if a table has only one key column, you can just pass one ``key`` value
          and ``get`` will find it
        - if you pass a list instead of a tuple, ``get`` will convert the list
          and still find the correct entry
        - if key is or consists of ``TableCell`` instances, ``get`` will use their
          respective ``.value``

        Returns ``None`` if entry is not found
        """
        if isinstance(key, list):
            key = tuple(key)
        if not isinstance(key, tuple):
            key = (key,)
        key = tuple(k.value if isinstance(k, self.CELL_TYPE) else k
                    for k in key)
        return self.index.get(key, None)

    def build_index(self):
        """(Re-)build the index of the rows based on the key-columns"""
        if len(self.definition.primary_key) == 0:
            raise ValueError(f"No primary key for {type(self.definition).__name__}")
        self.index = {}
        for row in self.rows:
            self.index[row.key()] = row

    def parse(self):
        retval = super().parse()
        if len(self.definition.primary_key) > 0:
            self.build_index()
        return retval

    def can_merge(self, other):
        """Whether or not the ``other`` table's content can be merged

        Conditions are:

        - no clashing entries in the primary keys
        - same table name
        """
        return self.definition.filename == other.definition.filename \
           and len(set(self.index.keys()) & set(other.index.keys())) == 0

    def merge(self, other):
        """Merge the ``other`` table's content into this table"""
        self.rows += other.rows
        self.rows.sort()
        if len(self.definition.primary_key) > 0:
            self.build_index()

    def sort(self, full=False):
        super().sort(full)
        if len(self.definition.primary_key) > 0:
            self.build_index()
        return self


class CSVDialect(csv.Dialect):
    """CSV dialect for SCOS-style ``.dat`` files"""
    delimiter = '\t'
    doublequote = False
    escapechar = None
    lineterminator = '\r\n'
    quoting = csv.QUOTE_NONE
    skipinitialspace = False
    strict = True


"""Framework to report errors in an SCOS instance"""
import enum
import sys


class FindingType(enum.Enum):
    INFO = 50
    WARNING = 40
    ERROR = 30


class Reporter:
    """Generic error report interface"""
    def error(self, table, line, description):
        """When an error was found in ``table``'s ``line``"""
        return self.report(FindingType.ERROR, table, line, description)

    def warning(self, table, line, description):
        """When a warning was found in ``table``'s ``line``"""
        return self.report(FindingType.WARNING, table, line, description)

    def info(self, table, line, description):
        """When an information statement should be reported about ``table``'s ``line``"""
        return self.report(FindingType.INFO, table, line, description)

    def report(self, finding, table, line, description):
        """A ``finding`` in ``table``'s ``line``"""
        raise NotImplementedError("Implement in a subclass")


class StdErrReporter(Reporter):
    """Simple report that just prints to stderr"""
    def report(self, finding, table, line, description):
        print(f"[{finding.name}] {table} Line {line}: {description}",
              file=sys.stderr)



"""The framework to write SCOS instance consistency/integrity checkers

To implement a new checker, subclass ``SCOSTableChecker`` and overwrite
at least the ``check_row`` function and the ``ACCEPTED_TABLES`` property to
select the tables that you wish to run this checker on.
"""
ALL_CHECKERS = []


class SCOSTableChecker:
    """The very basic visitor for SCOS table instances

    Subclasses are automatically registered in the global ``ALL_CHECKERS`` list
    to make it easier to just run all existing checkers on your SCOS instance.

    Parameters
    ----------
    scos : pyscos2000.SCOS2000
        is the instance of the SCOS-2000 that will be tested
    reporter : pyscos2000.Reporter
        the reporter to use for warnings, errors, etc.
    """
    ACCEPTED_TABLES = tuple()
    """The tables that this checker will be run on

    Must be an iterable of ``TableDefinition`` (not their instances).
    """

    def __init_subclass__(cls):
        ALL_CHECKERS.append(cls)

    def __init__(self, scos, reporter, pedantic=False):
        self.scos = scos
        self.reporter = reporter
        self.table = None
        self._cur_row = None
        self.pedantic = pedantic

    def visit(self, table):
        """Called when visiting a table"""
        if not self.accept(table):
            return

        self.table = table
        for row in table.rows:
            self._cur_row = row
            self.check_row(row)

    def accept(self, table):
        """Whether or not this table can be checked by this visitor

        The default implementation will use ``self.ACCEPTED_TABLES``, but you
        can overwrite this function to implement your own way of checking
        whether or not you want to run this checker on the given ``table``.

        Parameters
        ----------
        table : pyscos2000.Table
            Instance of a ``pyscos2000.instance.Table`` from the ``pyscos2000.SCOS``
            instance that's being checked
        """
        return isinstance(table.definition, self.ACCEPTED_TABLES)

    def check_row(self, row):
        """Called when a row is being visited"""

    def error(self, description):
        """Report an error in the current table and row"""
        assert self.reporter is not None
        assert self.table is not None
        assert self._cur_row is not None
        self.reporter.error(self.table.name,
                            self._cur_row.linenr,
                            description)

    def warning(self, description):
        """Report a warning in the current table and row"""
        assert self.reporter is not None
        assert self.table is not None
        assert self._cur_row is not None
        self.reporter.warning(self.table.name,
                              self._cur_row.linenr,
                              description)

    def info(self, description):
        """Report an informational statement in the current table and row"""
        assert self.reporter is not None
        assert self.table is not None
        assert self._cur_row is not None
        self.reporter.info(self.table.name,
                           self._cur_row.linenr,
                           description)



"""SCOS-2000 relational data model"""
from .schema import (
    IndicatorType, Endianess, OutputView, InterpolationBehaviour,
    ParameterNature, Justification, NumericFormatType, Radix,
    LimitInterpretation, LimitTypeInterpretation, MonitoringCheckFlag,
    )
from .shared import (
        unpack_uint_format, unpack_int_format, Error,
        UINT_BYTE_SIZES, INT_BYTE_SIZES
    )
from .parametertypes import ParameterType
from .synthetic import DynamicSyntheticParameter
from .concrete import Parameter, PacketInterface
from .checker import ALL_CHECKERS, SCOSTableChecker
from .reporter import Reporter, StdErrReporter
from .scos2000 import SCOS



"""Very simple tool to browse a SCOS-2000 database"""
import argparse
import re
from pathlib import Path

from pyscos2000 import SCOS


def main():
    args = parse_args()
    scos = SCOS(Path(args.database).expanduser())

    if args.check:
        scos.check()

    if args.tm is not None:
        tmpattern = re.compile(args.tm, re.I)
        for row in sorted(scos.pid.rows):
            description = row['PID_DESCR'].value
            spid = row['PID_SPID'].value
            text = f"{spid} {description}"
            if len(args.tm) > 0 and tmpattern.search(text) is None:
                continue
            print(f"({row['PID_TYPE'].value},{row['PID_STYPE'].value})",
                  spid, description)

    if args.fields is not None:
        tmpattern = re.compile(args.fields, re.I)
        for row in sorted(scos.pid.rows):
            description = row['PID_DESCR'].value
            spid = row['PID_SPID'].value
            text = f"{spid} {description}"
            if tmpattern.search(text) is None:
                continue
            print(spid)

            for plf in sorted(scos.plf.rows):
                if plf['PLF_SPID'].value != spid:
                    continue
                pcf = scos.pcf.get(plf['PLF_NAME'].value)
                print(pcf['PCF_NAME'].value, pcf['PCF_DESCR'].value)


def parse_args():
    parser = argparse.ArgumentParser()

    parser.add_argument('database',
                        type=str,
                        help='Location of the SCOS database to browse or inspect')

    parser.add_argument('--check',
                        action='store_true',
                        default=False,
                        help='Run all checkers on the database')

    parser.add_argument('-m', '--tm',
                        default=None,
                        type=str,
                        help="List all matching TM packets (accepts regex pattern)")

    parser.add_argument('-c', '--tc',
                        default=None,
                        type=str,
                        help="List all matching TC packets")

    parser.add_argument('-f', '--fields',
                        default=None,
                        nargs="?",
                        help="List the fields of this TM packet (accepts regex pattern)")

    parser.add_argument('-p', '--parameters',
                        default=None,
                        nargs="?",
                        help="List the parameters of this TC")

    return parser.parse_args()


if __name__ == '__main__':
    main()
